from shapely.geometry import mapping, MultiPolygon, Polygon as Poly, GeometryCollection
import numpy as np
import cv2

from lost_ds.geometry.api import Geometry
from lost_ds.vis.geometries import draw_polygons

class Polygon(Geometry):
    
    def __init__(self):
        super().__init__()


    def to_shapely(self, data):
        return Poly(data)
    
    
    def crop(self, crop_pos, data, **kwargs):
        intersection, intersects = self._crop_intersection(crop_pos, data)
        if not intersects:
            return intersection
        xmin, ymin, xmax, ymax = crop_pos.bounds
        new_polys = []
        if isinstance(intersection, MultiPolygon):
            new_polys = list(intersection.geoms)
        elif isinstance(intersection, GeometryCollection):
            new_polys = [i for i in intersection if 'polygon' in str(type(i))]
        else:
            new_polys = [intersection]
        
        for i, polygon in enumerate(new_polys):
            for p in mapping(polygon)['coordinates']:
                new_polys[i] = np.array(p).squeeze() - [xmin, ymin]
            
        new_polys = [p for p in new_polys if self.validate(p)]
        
        if not len(new_polys):
            return [np.nan]
        else:
            return new_polys
        
                        
    def validate(self, data):
        return len(data.shape)==2 and len(data)>=3


    def segmentation(self, segmentation, color, anno_data, anno_format, 
                     anno_style, **kwargs):
        anno_data = self.to_abs(anno_data, anno_format, segmentation.shape)
        cv2.fillPoly(segmentation, [anno_data.astype(np.int32)], color)
        return segmentation


    def _draw(self, img, data, style, text, color, line_thickness, fontscale, 
              **kwargs):
        if line_thickness is None:
            line_thickness = 2
        return draw_polygons(img, data, text, color, line_thickness, 
                             fontscale=fontscale)


import math

import numpy as np
import pandas as pd 
from shapely.geometry import box

from lost_ds.im_util import pad_image
from lost_ds.im_util import get_imagesize, get_fs
from lost_ds.geometry.lost_geom import LOSTGeometries
from lost_ds.io.file_man import FileMan


class DSCropper:
    """ Class to apply crops to images and annotations """
    def __init__(self, 
                 lost_geometry:LOSTGeometries=None, 
                 filesystem:FileMan=None) -> None:
        self.geom = lost_geometry
        if lost_geometry is None:
            self.geom = LOSTGeometries()
        self.fs = get_fs(filesystem)
        

    def crop_anno(self, img_path, df, crop_position, im_w=None, im_h=None, 
                  padding=None):
        """Calculate annos for a crop-position in an image
        
        Args:
            img_path (str): image to calculate the crop-annos
            crop_position (list): crop-position like: [xmin, ymin, xmax, ymax] 
                in absolute data-format
            im_w (int): width of image if available
            im_h (int): height of image if available
            df (pd.DataFrame): dataframe to apply 
            padding (tuple): image padding in absolute pixel values -
                ((top_pad, bot_pad), (left_pad, right_pad))
            
        Returns:
            pd.DataFrame 
        """
        if not im_w or not im_h:
            im_h, im_w = get_imagesize(img_path, self.fs)
        if padding is None:
            padding = ((0, 0), (0, 0))
        img_df = df[df['img_path']==img_path]
        assert max(crop_position) > 1, 'Crop position has to be absolute!'
        # crop_position = np.clip(crop_position, 0, [im_w, im_h, im_w, im_h])
        crop_box = box(*crop_position)
        result_annos = []
        for idx, row in img_df.iterrows():
            cropped_anno = self.geom.apply_crop(row.anno_data, row.anno_dtype, 
                                                row.anno_format, row.anno_style, 
                                                (im_h, im_w), crop_box, padding)
            for crop_anno in cropped_anno:
                new_anno = row.copy()
                new_anno['anno_data'] = crop_anno
                result_annos.append(new_anno)
                
        return pd.DataFrame(result_annos)
        

    def crop_img(self, img, crop_shape=(512, 512), overlap=(0, 0), 
                 fill_value=0):
        '''Crop a 3D image and get a 4D array with crops
        
        Args:
            img (np.array): (H, W, C) Image to crop
            crop_shape (tuple): (H, W) shape of crops
            overlap (tuple): (H, W) overlap between crops
            fill_value (float): pixel-value to fill the area where crops are 
                out of the image
        
        Returns:
            (np.ndarray, np.ndarray, tuple): 
                1. batch of cropped images as 4D-array with shape 
                    [N, crop_shape[0], crop_shape[1], channels] 
                2. positions of shape [N, 4] in format [xmin, ymin, xmax, ymax]
                3. image padding ((top_pad, bot_pad), (left_pad, right_pad))
        '''
        im_h, im_w, im_c = img.shape
        ol = overlap
        num_y_crops = math.ceil((im_h - ol[0]) / (crop_shape[0] - ol[0]))
        num_x_crops = math.ceil((im_w - ol[1]) / (crop_shape[1] - ol[1]))
        target_shape = [num_y_crops*(crop_shape[0] - ol[0]) + ol[0],
                        num_x_crops*(crop_shape[1] - ol[1]) + ol[1]]
        
        img, padding = pad_image(img, target_shape, fill_value)
        xy_steps = np.array(crop_shape) - ol
        crops = self._sliding_window_view(img,
                                    tuple(crop_shape) + (im_c, ),
                                    tuple(xy_steps) + (1, ))
        crop_pos = self._crop_positions(num_x_crops, num_y_crops, crop_shape,ol)
        
        return crops.reshape((-1, *crop_shape, im_c)), crop_pos, padding


    def _sliding_window_view(self, arr, window_shape, steps):
        ''' get sliding window view
        
        Args:
            arr (np.ndarray): array to slide over
            window_shape (tuple): shape of the window
            steps (tuple): steps of the window in each dimension
        
        Returns:
            np.ndarray: view of the defined sliding window
        '''
        # Code basically borrowed from
        # https://gist.github.com/meowklaski/4bda7c86c6168f3557657d5fb0b5395a
        in_shape = np.array(arr.shape[-len(steps):])
        window_shape = np.array(window_shape)
        steps = np.array(steps)
        nbytes = arr.strides[-1]  # size (bytes) of an element in `arr`
        window_strides = tuple(np.cumprod(arr.shape[:0:-1])[::-1]) + (1,)
        step_strides = tuple(window_strides[-len(steps):] * steps)
        strides = tuple(int(i) * nbytes for i in step_strides + window_strides)
        outshape = tuple((in_shape - window_shape) // steps + 1)
        outshape = outshape + arr.shape[:-len(steps)] + tuple(window_shape)
        return np.lib.stride_tricks.as_strided(arr, shape=outshape, 
                                               strides=strides, writeable=False)


    def _crop_positions(self, num_x_crops, num_y_crops, crop_shape, overlap):
        '''calculate crop positions for an image
        
        Args:
            num_x_crops (int): number of crops in x direction
            num_y_crops (int): number of crops in y direction
            crop_shape (tuple of int): (H, W)/(y, x) size of one crop
            overlap (tuple of int): (H, W)/(y, x) overlap between crops
            
        Returns:
            np.ndarray: crop positions in shape [N, 4] in style 
                [xmin, ymin, xmax, ymax]. 
                The order of the crops is from top left to bottom right. The 
                indices are to interpret like in this 3x3 example:
                 ___________
                |_0_|_1_|_2_|
                |_3_|_4_|_5_|
                |_6_|_7_|_8_|
                
        '''
        batch_size = num_x_crops * num_y_crops
        y_crop_i = np.sort(np.tile(np.arange(num_y_crops), num_x_crops))
        x_crop_i = np.tile(np.arange(num_x_crops), num_y_crops)
        # crop ids of shape [N, 4] with [xmin, ymin, xmax, ymax]
        crop_i = np.transpose([x_crop_i, y_crop_i, 1+x_crop_i, 1+y_crop_i]) 
        crop_positions = np.full((batch_size, 4), crop_shape[::-1] * 2)
        crop_positions = crop_positions * crop_i 
        ol_offset = np.hstack([crop_i[:, :2] * overlap[::-1]] * 2)
        crop_positions -= ol_offset
        return crop_positions
        


from typing import Union
import numpy as np
import pandas as pd 
from shapely.geometry import box, Polygon as Poly


from lost_ds.geometry.api import Point, Polygon, Line, Bbox

class LOSTGeometries:
    
    def __init__(self):
        self.point = Point()
        self.polygon = Polygon()
        self.bbox = Bbox()
        self.line = Line()
    
    def _get_geometry(self, dtype):
        if dtype == 'polygon':
            return self.polygon
        elif dtype == 'point':
            return self.point
        elif dtype == 'bbox':
            return self.bbox
        elif dtype == 'line':
            return self.line
    
    def serializable(self, data):
        if 'numpy' in str(type(data)):
            data = data.tolist()
        if not isinstance(data, list):
            return [[data]]
        elif isinstance(data, list) and len(data) == 0:
            return [data]
        elif not isinstance(data[0], list):
            return [data]
        elif not isinstance(data[0][0], list):
            return data           
        else:
            raise ValueError('Unsupported datatype: {}!'.format(type(data)))
    
    def validate(self, data, dtype):
        geom = self._get_geometry(dtype)
        if geom is not None:
            try:
                return geom.validate(data)
            except:
                return False
        else: 
            return True
    
    def poly_to_bbox(self, data, dst_style):
        '''Create a bbox from a polygon
        Args:
            data (pd.Series): polygon annodata
            dst_style (str): dst style for bbox
        Returns:
            pd.Series like row with bbox data
        '''
        return self.bbox.from_polygon(data, dst_style)
    
    
    def to_abs(self, data, dtype, data_format, img_shape):
        ''' Transform annotation to absolute
        
        Args:
            data (np.ndarray): data to transform
            dtype (str): anno dtype {'polygon', 'bbox', ...}
            data_format (str): format of data {'rel', 'abs'}
            img_shape (tuple or list of int): image shape like (H, W)
        
        Returns:
            np.ndarray: annotation transformed to absolute
        '''
        new_data = data
        if isinstance(data, np.ndarray):
            geom = self._get_geometry(dtype)
            new_data = geom.to_abs(data, data_format, img_shape)
        return new_data
    
    
    def to_rel(self, data, dtype, data_format, img_shape):
        ''' Transform annotation to relative
        
        Args:
            data (np.ndarray): data to transform
            dtype (str): anno dtype {'polygon', 'bbox', 'line', 'point'}
            data_format (str): format of data {'rel', 'abs'}
            img_shape (tuple or list of int): image shape like (H, W)
        
        Returns:
            np.ndarray: annotation transformed to relative
        '''
        new_data = data
        if isinstance(data, np.ndarray):
            geom = self._get_geometry(dtype)
            new_data = geom.to_rel(data, data_format, img_shape)
        return new_data
    
    
    def apply_crop(self, data, dtype, data_format, data_style, img_shape,
                   crop_position:Union[list, Poly], padding):
        ''' Crop an annotation
        
        Args:
            data (np.ndarray): anno-data to crop
            dtype (str): anno dtype {'polygon', 'bbox', 'line', 'point'}
            data_format (str): anno format {'rel', 'abs'}
            img_shape (list or tuple of int): img shape like (H, W)
            crop_position (list or shapely.Polygon): crop position in absolute 
                data format. If it is 
            padding (tuple): image padding in absolute pixel values -
                ((top_pad, bot_pad), (left_pad, right_pad))
        
        Returns:
            list of np.ndarray: cropped data in original data_format
        '''
        new_data = [data]
        if isinstance(data, np.ndarray):
            geom = self._get_geometry(dtype)
            if isinstance(crop_position, list):
                crop_position = box(*crop_position)
            tmp_format = 'abs'
            new_data = self.to_abs(data, dtype, data_format, img_shape)
            new_data = geom.offset(new_data, (padding[0][0], padding[1][0]))
            new_data = geom.apply_crop(crop_position, new_data, data_style)
            # reverse format to original
            if data_format == 'rel':
                xmin, ymin, xmax, ymax = crop_position.bounds
                crop_h, crop_w = ymax - ymin, xmax - xmin
                for i in range(len(new_data)):
                    new_data[i] = self.to_rel(new_data[i], dtype, tmp_format, 
                                              (crop_h, crop_w))
        return new_data
    
    
    def transform_bbox_style(self, data, src_style, dst_style):
        return self.bbox.transform_style(data, src_style, dst_style)
    
    
    #
    #   Drawing
    #
    def segmentation(self, segmentation, color, anno_data, anno_dtype, 
                     anno_format, anno_style, **kwargs):
        ''' draw annotation onto an image as segmentation
        
        Args:
            segmentation (np.ndarray): segmentation image to draw on
            color (int): color to use for segmentation
            anno_data (np.ndarray): annotation to draw
            anno_dtype (str): anno-type {'point', 'polygon', 'bbox', 'line'}
            anno_format (str): data format {'rel', 'abs'}
            anno_style (str): anno-style {'xy', 'xywh', 'xcycwh', 'x1y1x2y2'}

        Returns:
            image with segmentation drawn
        '''
        geom = self._get_geometry(anno_dtype)
        segmentation = geom.segmentation(segmentation, color, anno_data, 
                                         anno_format, anno_style, **kwargs)
        return segmentation
    
    
    def draw(self, img, anno_data, anno_conf, anno_label, anno_dtype, 
             anno_style, anno_format, line_thickness, fontscale, color, radius):
        '''draw annotation onto an image
        Args:
            img (np.ndarray): image to draw on
            anno_data (list of np.ndarray): annotations to draw
            anno_conf (list of float, list of list of float): confidences
            anno_label (list of str, list of list of str): labels
            anno_dtype (list of string): dtypes {'bbox', 'polygon', ...}
            anno_style (list of string): style {'xy', 'xywh', ...}
            anno_format (list of string): format {'abs', 'rel'}
            line_thickness (int, dict): int or dtype mapping like {dtype: int}.
                Value of -1 does fill the annotation.
            color (tuple, dict): (B,G,R) or class mapping like {class: (B,G,R)}
            radius (int): radius of circles/points
        Returns:
            np.ndarray: image with drawn annotations
        '''
        for i, data in enumerate(anno_data):
            if data is None:
                continue
            dtype = anno_dtype[i]
            style = anno_style[i] 
            format_ = anno_format[i]
            lbl = anno_label[i]
            if isinstance(color, list):
                c = color[i]
            else:
                c = color
            conf = None
            if isinstance(anno_conf, list):
                conf = anno_conf[i]
            text = self.get_text(lbl, conf)
            c = self.get_color(lbl, c)
            thickness = self.get_line_thickness(dtype, line_thickness)    
            geom = self._get_geometry(dtype)
            img = geom.draw(img, data, format_, style, text, c, thickness, 
                            fontscale, radius=radius)
        return img
        
        
    def get_text(self, label, confidence):
        '''
        Args:
            label (string, list of string): single label or multi label
            confidence (float or list of float):
        '''
        if np.array([pd.isna(label)]).all():
            return None
        conf = text = None
        lbl = label
        if isinstance(label, str):
            if len(label) == 0:
                return None
            lbl = [label]
        if len(lbl) == 0:
            return None
        if isinstance(confidence, float) and pd.notna(confidence):
            conf = [confidence]
        if pd.notna(conf):
            text = [l+' ('+str(round(c, 2))+')' for c, l in zip(conf, lbl)]
        else:
            text = [l for l in lbl]
        text = ', '.join(text)
        return text
    
    
    def get_color(self, label, color):
        '''Decode color for class
        Args:
            label (string, list of string): single label or multi label
            color (tuple, dict): tuple with color or dict with mapping of colors
                for different classes
        Returns:
            tuple color (B,G,R)
        '''
        if isinstance(color, tuple):
            return color
        elif isinstance(color, dict):
            lbl = label
            # just pick the first one in case of multilabel
            if isinstance(label, list):
                lbl = label[0]
            lbl = lbl[0]        # ???
            if lbl in color.keys():
                return color[lbl]
            else:
                return None
        
        
    def get_line_thickness(self, dtype, thickness):
        '''Decode thickness for anno type
        Args:
            dtype (string): annotype 
            thickness (int, dict): thickness or dict with mapping of thickness
                for different annotypes
        Returns:
            int thickness
        '''
        if isinstance(thickness, int):
            return thickness
        elif isinstance(thickness, dict):
            if dtype in thickness.keys():
                return thickness[dtype]
            else: 
                return None
        


import cv2
import numpy as np
import cv2

    
#
#   draw text
#


def get_textsize(text, thickness, fontscale):
    (text_w, text_h), baseline = cv2.getTextSize(text,
                                                cv2.FONT_HERSHEY_SIMPLEX,
                                                fontscale, 
                                                thickness)
    return text_w, text_h, baseline


def draw_text(img, text, x, y, color, thickness=2, fontscale=2):
    '''Draw text onto an image
    Args:
        img (numpy.ndarray): Image to draw on
        text (str): Text to draw
        x (int): x-coordinate of geometries point where to draw the text
        y (int): y-coordinate of geometries point where to draw the text
        color (tuple): BGR color to use
        thickness (int): Text and line thickness
    Returns:
        np.ndarray Image with text drawn and framed
    Note:
        The Text will be white (255,255,255), on black background, framed by
        a rectangle having the specified color.
    '''
    if text is None:
        return img
    
    text_w, text_h, baseline = get_textsize(text, thickness, fontscale)
    
    # avoid text being out of image
    ymin = y - text_h - baseline - thickness
    if ymin < 0:
        y += abs(ymin)
        ymin = thickness
    x -= thickness
    xmax = x + text_w
    if xmax > img.shape[1]:
        x -= xmax - img.shape[1] - thickness
        xmax = img.shape[1] - thickness
        
    # black filled rectangle
    cv2.rectangle(img, (x, ymin), (xmax, y), (0,0,0),
                  thickness=cv2.FILLED)
    # colored rectangle frame to get a colored frame to the black rect
    cv2.rectangle(img, (x-thickness, ymin-thickness),
                  (xmax + thickness, y + thickness),
                  color, thickness=thickness)
    # white text on black rectangle
    cv2.putText(img, text, (x, y-baseline//2), cv2.FONT_HERSHEY_SIMPLEX, 
                fontscale, (255, 255, 255), thickness)
    
    return img



#
#   Draw geometries
#


def get_data(data, text, color):
    if color is None:
        color = (255, 255, 255)    
    if not isinstance(data, list):
        data = [data.astype(np.int32)]
    if not isinstance(text, list):
        text = [text]
    if not isinstance(color, list):
        color = [color] * len(data)
    
    assert len(data) == len(text) == len(color)
    
    return data, text, color


def draw_points(img, data, text, color=(0, 0, 255), radius=2, 
                line_thickness=-1, fontscale=2):
    '''Draw points onto the given image

    Args:
        img (str or array): Path to an image file or an image as numpy rgb array.
        data (np.ndarray, list of np.ndarray): points in format [x, y] 
            Values have to be absolute.
        text (str, list of str): text to draw onto the annotato
        color (tuple, list of tuple): BGR color to use
        line_thickness (int): Thickness of Point

    Returns:
        np.array: Image in opencv style
    '''
    data, text, color = get_data(data, text, color)
    
    if len(img.shape) < 3:
        img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
        
    for point, txt, c in zip(data, text, color):
        cv2.circle(img, tuple(point), radius, c, line_thickness)
        img = draw_text(img, txt, *point-[0, radius], c, max(line_thickness, 1), 
                        fontscale)
    return img


def draw_polygons(img, data, text, color=(0, 0, 255), line_thickness=2, 
                  fontscale=2, offset=None):
    '''Draw polygons onto the given image

    Args:
        img (np.ndarray): image as numpy array.
        data (np.ndarray, list of np.ndarray): polygons format 
            [[x,y],[x,y],[x,y],...]. Values have to be absolute.
        text (string, list of strings): text to draw onto the annotation
        conf_list (list): List of confidences for each annotation.
        color (tuple, list of tuple): BGR colors to use
        line_thickness (int): Thickness of the polygon lines.

    Returns:
        np.array: Image in opencv style
    '''
    data, text, color = get_data(data, text, color)
    
    if len(img.shape) < 3:
        img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
        
    for poly, txt, c in zip(data, text, color):
        if offset is not None and isinstance(offset, (list, tuple, np.ndarray)):
            poly = poly + offset
        y_values = poly[:, 1]
        txt_y_i = np.argmin(y_values)
        txt_x = poly[txt_y_i][0]
        txt_y = y_values.min()
        cv2.polylines(img, [poly], True, c, thickness=line_thickness)
        img = draw_text(img, txt, txt_x, txt_y, c, line_thickness, fontscale)
    return img


def draw_boxes(img, data, text, color=(0, 0, 255), line_thickness=1,
               fontscale=2):
    '''Draw bboxes onto the given image

    Args:
        img (np.ndarray): image as numpy array
        data (np.ndarray, list of np.ndarray): bboxes in format, 
            [xmin, ymin, xmax, ymax]. Values have to be absolute.
        text (str, list of str): text to draw onto the annotation, e.g. label 
            of confidence 
        color (tuple, list of tuples): BGR color that is used for all boxes.
        line_thickness (int): Thickness of the bounding box lines.
        
    Returns:
        np.array: Image with bboxes drawn
    '''
    
    data, text, color = get_data(data, text, color)
    
    if len(img.shape) < 3:
        img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
    
    for bb, txt, c in zip(data, text, color):
        cv2.rectangle(img, (bb[0],bb[1]), (bb[2],bb[3]), c, line_thickness)
        img = draw_text(img, txt, bb[0], bb[1], c, line_thickness, fontscale)
    return img


def draw_lines(img, data, text, color=(0, 0, 255), line_thickness=2,
               fontscale=2):
    '''Draw poly lines onto the given image

    Args:
        img (np.ndarray): image as numpy array.
        data (np.ndarray, list of np.ndarray): lines format 
            [[x,y],[x,y],[x,y],...]. Values have to be absolute.
        text (string, list of strings): text to draw onto the annotation
        conf_list (list): List of confidences for each annotation.
        color (tuple, list of tuple): BGR colors to use
        line_thickness (int): Thickness of the lines.

    Returns:
        np.array: Image in opencv style
    '''
    data, text, color = get_data(data, text, color)
    
    if len(img.shape) < 3:
        img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
        
    for poly, txt, c in zip(data, text, color):
        y_values = poly[:, 1]
        txt_y_i = np.argmin(y_values)
        txt_x = poly[txt_y_i][0]
        txt_y = y_values.min()
        cv2.polylines(img, [poly], False, c, line_thickness)
        img = draw_text(img, txt, txt_x, txt_y, c, line_thickness, fontscale)
    return img
    


import math

import numpy as np 

from lost_ds.util import get_fs
    
def get_imagesize(path, filesystem=None):
    '''Get the size of an image without reading the entire data
    
    Args:
        path (str): path to image
        filesystem (fsspec.filesystem, FileMan): filesystem to use. Use local
            if not initialized
        
    Returns:
        tuple of int: (im_h, im_w) height and width of the image
    '''
    fs = get_fs(filesystem)
    return fs.get_imagesize(path)


def pad_image(img, target_shape, fill_value):
    ''' Pad image to target shape
    
    Args:
        img (np.ndarray): image to pad
        target_shape (tuple): desired shape after padding [H, W]
        fill_value (int): Value to use for padding
        
    Returns:
        tuple(np.ndarray, tuple): 
            np.ndarray - padded image with target_shape
            tuple - padding ((top_pad, bot_pad), (left_pad, right_pad))
    '''
    d_pad_h = target_shape[0] - img.shape[0]
    d_pad_w = target_shape[1] - img.shape[1]
    top_pad = math.floor(d_pad_h / 2)
    bot_pad = math.ceil(d_pad_h / 2)
    left_pad = math.floor(d_pad_w / 2)
    right_pad = math.ceil(d_pad_w / 2)
    # do not pad in depth dimension (vertical_pad, horizontal_pad, depth_pad)
    pads = ((top_pad, bot_pad), (left_pad, right_pad), (0, 0))
    return np.pad(img, pads, constant_values=fill_value), pads[:2]



import os

from tqdm import tqdm
import pandas as pd 
from joblib import Parallel, delayed, cpu_count
from lost_ds.functional.filter import label_selection, selection_mask, remove_empty
import numpy as np
from lost_ds.functional.transform import to_abs, polygon_to_bbox

from lost_ds.io.file_man import FileMan
from lost_ds.functional.validation import validate_empty_images
from lost_ds.cropping.ds_cropper import DSCropper
from lost_ds.util import get_fs
from lost_ds.im_util import get_imagesize


def crop_anno(img_path, crop_position, df, im_w=None, im_h=None, 
              filesystem:FileMan=None):
    """Calculate annos for a crop-position in an image
    
    Args:
        img_path (str): image to calculate the crop-annos
        crop_position (list): crop-position like: [xmin, ymin, xmax, ymax] 
            in absolute data-format
        im_w (int): width of image if available
        im_h (int): height of image if available
        df (pd.DataFrame): dataframe to apply 
        filesystem (fsspec.filesystem, FileMan): filesystem to use. Use local
            if not initialized
            
    Returns:
        pd.DataFrame 
    """
    cropper = DSCropper(filesystem=filesystem)
    return cropper.crop_anno(img_path, df, crop_position, im_w, im_h)

def crop_img(img, crop_shape=(512, 512), overlap=(0, 0), 
                fill_value=0, filesystem:FileMan=None):
        '''Crop a 3D image and get a 4D array with crops
        
        Args:
            img (np.array): (H, W, C) Image to crop
            crop_shape (tuple): (H, W) shape of crops
            overlap (tuple): (H, W) overlap between crops
            fill_value (float): pixel-value to fill the area where crops are 
                out of the image
        
        Returns:
            (np.ndarray, np.ndarray, tuple): 
                1. batch of cropped images as 4D-array with shape 
                    [N, crop_shape[0], crop_shape[1], channels] 
                2. positions of shape [N, 4] in format [xmin, ymin, xmax, ymax]
                3. image padding ((top_pad, bot_pad), (left_pad, right_pad))
        '''
        cropper = DSCropper(filesystem=filesystem)
        return cropper.crop_img(img, crop_shape=crop_shape, overlap=overlap, 
                                fill_value=fill_value)
    
    
def crop_dataset(df, dst_dir, crop_shape=(500, 500), overlap=(0,0),
                 write_empty=False, fill_value=0, filesystem:FileMan=None,
                 parallel=-1):
    """Crop the entire dataset with fixed crop-shape

    Args:
        df (pd.DataFrame): dataframe to apply bbox typecast
        dst_dir (str): Directory to store the new dataset
        crop_shape (tuple, list): [H, W] cropped image dimensions
        overlap (tuple, list): [H, W] overlap between crops
        write_empty (bool): Flag if crops without annos wil be written
        fill_value (float): pixel value to fill the rest of crops at borders
        filesystem (fsspec.filesystem, FileMan): filesystem to use. Use local
            if not initialized
    Returns:
        pd.DataFrame
    """
    fs = get_fs(filesystem)
    fs.makedirs(dst_dir, True)
    df = validate_empty_images(df)
    
    def crop_and_recalculate(img_path, img_df):
        cropper = DSCropper(filesystem=filesystem)
        img = cropper.fs.read_img(img_path)
        im_h, im_w, im_c = img.shape
                
        crops, positions, padding = cropper.crop_img(img, crop_shape, overlap, 
                                                     fill_value)
        img_name, img_ending = img_path.split('/')[-1].split('.')
        
        result_df = []
        for i, position in enumerate(positions):
            crop_string = f'_shape_{crop_shape[0]}_{crop_shape[1]}'
            overlap_string = f'_overlap_{overlap[0]}_{overlap[1]}'
            crop_name = img_name + '_crop_' + str(i) + crop_string + overlap_string + '.' + img_ending
            crop_path = os.path.abspath(os.path.join(dst_dir, crop_name))
            crop_df = cropper.crop_anno(img_path, img_df, position, im_w, im_h, 
                                        padding)
            data_present = crop_df['anno_data'].notnull()
            
            # padding is ((top, bot), (left, right))
            pad_y, pad_x = padding[0][0], padding[1][0]
            crop_position = position - [pad_x, pad_y, pad_x, pad_y]
            
            if data_present.any() or write_empty:
                crop_df['img_path'] = crop_path
                crop_df['crop_position'] = crop_df['img_path'].map(lambda x: crop_position)
                cropper.fs.write_img(crops[i], crop_path)
                data_df = crop_df[data_present]
                if write_empty:
                    empty_df = crop_df.loc[~data_present].drop_duplicates('img_path')
                    if not len(data_df):
                        data_df = pd.concat([data_df, empty_df])
                result_df.append(data_df)
                
        if len(result_df):
            return pd.concat(result_df)
        else:
            return None
    
    if parallel:
        crop_dfs = Parallel(n_jobs=parallel)(delayed(crop_and_recalculate)(path, img_df) 
                                        for path, img_df in tqdm(df.groupby('img_path'), 
                                                        desc='crop dataset'))
    else:
        crop_dfs = []
        for path, img_df in tqdm(df.groupby('img_path'), desc='crop dataset'):
            crop_dfs.append(crop_and_recalculate(path, img_df))
    
    ret_df = pd.concat(crop_dfs)
    
    anno_keys = [k for k in ret_df.keys() if 'anno' in k]
    # set anno-information from formerly non empty annos to empty
    ret_df.loc[ret_df['anno_data'].isnull(), anno_keys] = None
    # make anno_lbl* format compatible
    anno_keys = [k for k in ret_df.keys() if 'anno_lbl' in k]
    for k in anno_keys:
        ret_df.loc[ret_df['anno_data'].isnull(), k] = ret_df[ret_df['anno_data'].isnull()][k].map(lambda x: np.array([], 'object'))

    return ret_df


# def crop_components(df, dst_dir, base_labels=-1, lbl_col='anno_lbl', context=0, 
#                     context_alignment=None, min_size=None, 
#                     anno_dtype=['polygon'], center_lbl_key='center_lbl', 
#                     filesystem:FileMan=None, parallel=-1):
#     """Crop the entire dataset with fixed crop-shape

#     Args:
#         df (pd.DataFrame): dataframe to apply bbox typecast
#         dst_dir (str): Directory to store the new dataset
#         base_labels (list of str): labels to align the crops 
#         lbl_col (str): column holding the labels
#         context (float, tuple of floats): context to add to each component for 
#             cropping (twice -> left/right, top/bottom). If tuple of float: uses 
#             the given floats as fraction of the components (H, W) to calculate 
#             the added contexts. 
#         context_alignment (str): Define alignment of the crop-context. One of 
#             [None, 'max', 'min', 'flip']. 
#             None: Use H/W-context in H/W-dimension
#             max: use maximum of H-context, W-context for both dimensions
#             min: use minimum of H-context, W-context for both dimensions
#             flip: use H-context in W-dimension and vice versa. This makes the 
#                 crop a little bit more squarish shaped
#         min_size (int, tuple of int): minimum size of produced crops in both
#             dimensions if int or for (H, W) if tuple of int
#         anno_dtype (list of str): dtype to apply on
#         center_lbl_key (str): column containing the label the crop was aligned to
#         filesystem (fsspec.filesystem, FileMan): filesystem to use. Use local
#             if not initialized
            
#     Returns:
#         pd.DataFrame
#     """
    
#     cropper = DSCropper(filesystem=filesystem)
#     fs = get_fs(filesystem)
#     fs.makedirs(dst_dir, True)
#     df = validate_empty_images(df)
#     if len(np.setdiff1d(anno_dtype, ['polygon', 'bbox'])):
#         raise NotImplementedError(f'Component cropping does not support {anno_dtype}')
    
#     df = to_abs(df, filesystem=filesystem, verbose=False)
#     df[center_lbl_key] = False
#     context_y = context_x = context
#     if not isinstance(context, (float, int)):
#         context_y, context_x = context
    
#     min_size_y = min_size_x = 0
#     if not isinstance(min_size, int) and min_size is not None:
#         min_size_y, min_size_x = min_size
#     if isinstance(min_size, int):
#         min_size_x = min_size_y = min_size
    
#     def crop_and_recalculate(img_path, img_df):
#         if base_labels == -1:
#             base_df = img_df
#         else:
#             base_df = label_selection(base_labels, img_df, col=lbl_col)
#         if not len(base_df):
#             return None
#         crop_df = base_df[selection_mask(anno_dtype, base_df, 'anno_dtype')].copy()
#         if not len(crop_df):
#             return None
#         crop_df = polygon_to_bbox(crop_df, 'x1y1x2y2')
        
#         img = fs.read_img(img_path)
#         im_h, im_w = img.shape[:2]
#         ret_df = list()
        
#         crop_id = 0
#         for idx, row in crop_df.iterrows():
#             minx, miny, maxx, maxy = row.anno_data
#             w, h = maxx - minx, maxy - miny
#             x_marg, y_marg = context_x * w, context_y * h
#             if context_alignment == 'flip':
#                 x_marg, y_marg = y_marg, x_marg
#             elif context_alignment == 'max':
#                 x_marg = y_marg = max(x_marg, y_marg)
#             elif context_alignment == 'min':
#                 x_marg = y_marg = min(x_marg, y_marg)
                
#             range_w = int(minx - x_marg), int(maxx + x_marg)
#             range_h = int(miny - y_marg), int(maxy + y_marg)
#             w, h = range_w[1] - range_w[0], range_h[1] - range_h[0]
            
#             if w < min_size_x:
#                 diff = min_size_x - w
#                 range_w = (int(range_w[0]-diff/2), int(range_w[1] + diff/2))
#             if h < min_size_y:
#                 diff = min_size_y - h
#                 range_h = (int(range_h[0]-diff/2), int(range_h[1] + diff/2))
            
#             minx, maxx,  = np.clip(range_w, 0, im_w) 
#             miny, maxy = np.clip(range_h, 0, im_h)
            
#             crop_pos = [minx, miny, maxx, maxy]
#             crop_anno = cropper.crop_anno(img_path, img_df, crop_pos, im_w, im_h)

#             # apply crop
#             crop = img[miny: maxy, minx: maxx]
#             crop_name = f'{img_path.split("/")[-1].split(".")[0]}_crop_{crop_id}.{img_path.split(".")[-1]}'
#             crop_id += 1
#             crop_path = os.path.join(dst_dir, crop_name)
#             fs.write_img(crop, crop_path)
#             crop_anno['img_path'] = crop_path
#             crop_anno['crop_position'] = crop_anno['img_path'].map(lambda x: [minx, miny, maxx, maxy])
#             crop_anno.loc[idx, center_lbl_key] = True
#             ret_df.append(crop_anno)    
#         if len(ret_df):
#             return pd.concat(ret_df)
#         else: 
#             return None
        
#     if parallel:
#         crop_dfs = Parallel(n_jobs=parallel)(delayed(crop_and_recalculate)(path, df) 
#                                              for path, df in tqdm(df.groupby('img_path'), 
#                                                                   desc='crop dataset'))
#     else:
#         crop_dfs = []
#         for path, df in tqdm(df.groupby('img_path'), desc='crop dataset'):
#             crop_dfs.append(crop_and_recalculate(path, df))
    
#     # TODO: return empty df if noting to crop
#     return pd.concat(crop_dfs)



def crop_components(df, dst_dir=None, base_labels=-1, lbl_col='anno_lbl', context=0, 
                    context_alignment=None, min_size=None, 
                    anno_dtype=['polygon'], center_lbl_key='center_lbl', 
                    filesystem:FileMan=None, parallel=-1):
    """Crop the entire dataset with fixed crop-shape

    Args:
        df (pd.DataFrame): dataframe to apply bbox typecast
        dst_dir (str, None): Directory to store the crops. If None no cropping 
            will be done.
        base_labels (list of str): labels to align the crops 
        lbl_col (str): column holding the labels
        context (float, tuple of floats): context to add to each component for 
            cropping (twice -> left/right, top/bottom). If tuple of float: uses 
            the given floats as fraction of the components (H, W) to calculate 
            the added contexts. 
        context_alignment (str): Define alignment of the crop-context. One of 
            [None, 'max', 'min', 'flip']. 
            None: Use H/W-context in H/W-dimension
            max: use maximum of H-context, W-context for both dimensions
            min: use minimum of H-context, W-context for both dimensions
            flip: use H-context in W-dimension and vice versa. This makes the 
                crop a little bit more squarish shaped
        min_size (int, tuple of int): minimum size of produced crops in both
            dimensions if int or for (H, W) if tuple of int
        anno_dtype (list of str): dtype to apply on
        center_lbl_key (str): column containing the label the crop was aligned to
        filesystem (fsspec.filesystem, FileMan): filesystem to use. Use local
            if not initialized
            
    Returns:
        pd.DataFrame
    """
     
    cropper = DSCropper(filesystem=filesystem)
    fs = get_fs(filesystem)
    crop_image = True
    if dst_dir is None:
        crop_image = False
    if crop_image:
        fs.makedirs(dst_dir, True)
    df = validate_empty_images(df)
    if len(np.setdiff1d(anno_dtype, ['polygon', 'bbox'])):
        raise NotImplementedError(f'Component cropping does not support {anno_dtype}')
    
    df = to_abs(df, filesystem=filesystem, verbose=False)
    df[center_lbl_key] = False
    context_y = context_x = context
    if not isinstance(context, (float, int)):
        context_y, context_x = context
    
    min_size_y = min_size_x = 0
    if not isinstance(min_size, int) and min_size is not None:
        min_size_y, min_size_x = min_size
    if isinstance(min_size, int):
        min_size_x = min_size_y = min_size
    
    def crop_and_recalculate(img_path, img_df):
        # pick labels to crop
        if base_labels == -1:
            base_df = img_df
        else:
            base_df = label_selection(base_labels, img_df, col=lbl_col)
        
        if not len(base_df):
            return None
        
        # pick anno types to crop
        crop_df = base_df[selection_mask(anno_dtype, base_df, 'anno_dtype')].copy()
        
        if not len(crop_df):
            return None
        
        # crops have to be bboxes
        crop_df = polygon_to_bbox(crop_df, 'x1y1x2y2')
        
        # get image and/or information
        if crop_image:
            img = fs.read_img(img_path)
            im_h, im_w = img.shape[:2]
        else:
            im_h, im_w = get_imagesize(img_path)
        
        # iterate over images, calculate crop, crop image and anno
        ret_df = list()
        crop_id = 0
        for idx, row in crop_df.iterrows():
            
            # get crop position
            minx, miny, maxx, maxy = row.anno_data
            w, h = maxx - minx, maxy - miny
            
            # calculate context and alignment as specified in args
            x_marg, y_marg = context_x * w, context_y * h
            if context_alignment == 'flip':
                x_marg, y_marg = y_marg, x_marg
            elif context_alignment == 'max':
                x_marg = y_marg = max(x_marg, y_marg)
            elif context_alignment == 'min':
                x_marg = y_marg = min(x_marg, y_marg)
            
            # new crop shape
            range_w = int(minx - x_marg), int(maxx + x_marg)
            range_h = int(miny - y_marg), int(maxy + y_marg)
            w, h = range_w[1] - range_w[0], range_h[1] - range_h[0]
            
            # crop min size
            if w < min_size_x:
                diff = min_size_x - w
                range_w = (int(range_w[0]-diff/2), int(range_w[1] + diff/2))
            if h < min_size_y:
                diff = min_size_y - h
                range_h = (int(range_h[0]-diff/2), int(range_h[1] + diff/2))
            
            # take care the crops are inside the image
            minx, maxx,  = np.clip(range_w, 0, im_w) 
            miny, maxy = np.clip(range_h, 0, im_h)
            
            # final crop position with aligned context
            crop_pos = [minx, miny, maxx, maxy]
            
            # crop annotation
            crop_anno = cropper.crop_anno(img_path, img_df, crop_pos, im_w, im_h)
            crop_anno.loc[idx, center_lbl_key] = True
            crop_anno['crop_position'] = crop_anno['img_path'].map(lambda x: [minx, miny, maxx, maxy])
            crop_anno['crop_id'] = crop_id
            
            # crop image and save it, add path to dataframe
            if crop_image:
                crop = img[miny: maxy, minx: maxx]
                crop_name = f'{img_path.split("/")[-1].split(".")[0]}_crop_{crop_id}.{img_path.split(".")[-1]}'
                crop_path = os.path.join(dst_dir, crop_name)
                fs.write_img(crop, crop_path)
                crop_anno['img_path'] = crop_path
            
            crop_id += 1
            
            # add crop dataframe as result
            ret_df.append(crop_anno)    
            
        # return results or None if no crops occured
        if len(ret_df):
            return pd.concat(ret_df)
        else: 
            return None
        
    if parallel:
        crop_dfs = Parallel(n_jobs=parallel)(delayed(crop_and_recalculate)(path, df) 
                                             for path, df in tqdm(df.groupby('img_path'), 
                                                                  desc='crop dataset'))
    else:
        crop_dfs = []
        for path, df in tqdm(df.groupby('img_path'), desc='crop dataset'):
            crop_dfs.append(crop_and_recalculate(path, df))
    
    # TODO: return empty df if nothing to crop
    return pd.concat(crop_dfs)



import fsspec, cv2
import shutil
import pandas as pd 
import numpy as np
from PIL import Image
    
    
class FileMan(object):
    
    '''File manager class to read and write datasets and images'''
    
    def __init__(self, filesystem=None, backend='pandas'):
        self.fs = self._get_filesystem(filesystem)    
        self.backend = self._get_backend(backend)
    
    def _get_filesystem(self, filesystem):
        fs = None
        if filesystem is None:
            fs = fsspec.filesystem('file')
        elif 'fsspec' in str(filesystem):
            fs = filesystem
        elif 'FileMan' in str(filesystem):
            fs = filesystem.fs
        else:
            fs = filesystem #For other fsspec filesystems 
        return fs 
    
    def _get_backend(self, backend):
        lib = None
        if backend == 'pandas':
            lib = pd 
        else:
            raise ValueError('Unsupported backend: {}!'.format(backend))
        return lib
    
    
    #
    #   fs utils
    #
    
    def exists(self, path):
        return self.fs.exists(path)
    
    def isdir(self, path):
        return self.fs.isdir(path)
    
    def isfile(self, path):
        return self.fs.isfile(path)
    
    def listdir(self, path):
        return self.fs.listdir(path)
    
    def ls(self, path):
        return self.fs.ls(path)
    
    def rm(self, path, recursive=False):
        self.fs.rm(path, recursive)
    
    def makedirs(self, path, exist_ok=True):
        return self.fs.makedirs(path, exist_ok=exist_ok)

    def copy(self, src_path, dst_path, recursive=False):
        try:
            self.fs.copy(src_path, dst_path, recursive)
        except shutil.SameFileError:
            pass
        
    
    def glob(self, path):
        return self.fs.glob(path)
    
    
    #
    #   open files
    #
    
    def open_text(self, path, mode='r'):
        return self.fs.open(path, mode)
    
    
    def open_binary(self, path, mode='r'):
        return self.fs.open(path, mode+'b')


    #
    #   R/W Images
    #
            
    def read_img(self, img_path, color_type='color'):
        if color_type == 'color':
            color = cv2.IMREAD_COLOR
        else:
            color = cv2.IMREAD_GRAYSCALE
        with self.open_binary(img_path, 'r') as f:
            arr = np.asarray((bytearray(f.read())), dtype=np.uint8)
            img = cv2.imdecode(arr, color)
        return img
    
    
    def write_img(self, img, img_path):
        success = False
        with self.open_binary(img_path, 'w') as f:
            success, arr = cv2.imencode('.' + img_path.split('.')[-1], img)
            f.write(bytearray(arr))
        return success
    
    
    def get_imagesize(self, path):
        '''Get the size of an image without reading the entire data
        
        Args:
            path (str, bytes): path or 'fs.open(path)'
            
        Returns:
            tuple of int: (im_h, im_w)
        '''
        im_h, im_w = None, None
        with self.open_binary(path) as f:
            im_w, im_h = Image.open(f).size
        return im_h, im_w

    
    #
    #   R/W Datasets
    #
    
    def _get_dstype(self, path):
        supported_types = ['csv', 'parquet']
        for t in supported_types:
            if t in path: 
                return t
        raise ValueError('Datatype of {} not supported!'.format(path))
    
    
    def load_dataset(self, ds):
        
        if isinstance(ds, str):
            return self._load_dataset(ds)
        elif isinstance(ds, dict):
            return pd.DataFrame(ds)
        elif isinstance(ds, pd.DataFrame):
            return ds
        elif 'LOSTDataset' in str(type(ds)):
            return ds.df
        elif ds is None:
            return None
        elif isinstance(ds, (list, tuple)):
            return pd.concat([self.load_dataset(df) for df in ds])
        else:
            raise ValueError('Unkown input-type {}'.format(type(ds)))
            
            
    def _load_dataset(self, ds_path):
        ds_type = self._get_dstype(ds_path)
        ds = None
        read_fn = getattr(self.backend, 'read_' + ds_type)
        try:
            ds = read_fn(ds_path)
        except: 
            open_fn = self.open_text if ds_type == 'csv' else self.open_binary
            with open_fn(ds_path, 'r') as f:
                ds = read_fn(f)            
        return ds


    def write_dataset(self, ds, ds_path):
        ds_type = self._get_dstype(ds_path)
        open_fn = self.open_text if ds_type == 'csv' else self.open_binary
        write_fn = getattr(ds, 'to_' + ds_type)
        with open_fn(ds_path, 'w') as f:
            write_fn(f, index=False)





# def create_instance_segmentation(self, dst_dir, fill_value=0, df=None, inplace=False):
#     '''Create pixelmaps from polygon-annos
    
#     Args:
#         dst_dir (str): Directory to store the pixel maps. 
#         fill_value (int): Pixel value for not annotated areas
#         df (pd.DataFrame): Frame to generate the pixel maps from
#         inplace (bool): Overwrite the own LOST-dataframe when True
    
#     Returns:
#         tuple: (train_df, big_df) The sem seg dataframe with unique paths
#             and the big dataframe with 'mask.path' column added
#     '''

#     os.makedirs(dst_dir, exist_ok=True)

#     if df is None:
#         df = self.df
    
#     df = self.remove_empty(df=df)
#     df = self.remove_empty_labels(df=df)

#     labels = self.get_unique_labels(df=df)
                
#     df = df[df['anno.dtype']=='polygon']
#     # REPLACE pack/unpack:
#     # df['anno.data'] = df['anno.data'].map(lambda x: json.loads(x))
#     # REPLACE pack/unpack:
#     # df['anno.lbl.name'] = df['anno.lbl.name'].map(lambda x: json.loads(x)[0])
#     df['anno.lbl.name'] = df['anno.lbl.name'].map(lambda x: x[0])
#     # TODO: rest of func with x instead of x[0] (label in list)
#     df['mask.path'] = None
    
#     px_values = {'anno.idx':[], 'mask.px_value':[]}

#     image_paths = self.get_unique_img_paths(df=df)
    
#     for i, image_path in enumerate(image_paths):    
#         print_progress_bar(i+1, len(image_paths), prefix='create px-maps:')
#         if not os.path.exists(image_path):
#             raise Exception('Image {} does not exist'.format(image_path))
#         im_w, im_h = get_image_size(image_path)
#         px_map = np.full([im_h, im_w], fill_value, dtype=np.int32)
#         image_annos = df[df['img.img_path']==image_path]
#         px_val = 1
#         for label in labels:
#             # get annos from current image
#             draw = image_annos[image_annos['anno.lbl.name']==label]
#             for poly, anno_id in zip(draw['anno.data'].to_list(), draw['anno.idx'].to_list()):
#                 # draw all annos
#                 poly = (np.array([list(point.values()) for point in poly]) * [im_w, im_h]).astype(np.int32)
#                 cv2.fillPoly(px_map, [poly], px_val)
#                 px_values['mask.px_value'].append(px_val)
#                 px_values['anno.idx'].append(anno_id)
#                 px_val += 1

#         img_name = image_path.split('/')[-1]
#         filename = img_name.split('.')[0]
#         filename = filename + '.png'
#         mask_path = os.path.join(dst_dir, filename)
#         cv2.imwrite(mask_path, px_map)
#         image_annos.loc[:,'mask.path'] = mask_path
#         df[df['img.img_path']==image_path] = image_annos

#     px_mapping = pd.DataFrame(px_values)
#     df = df.merge(px_mapping, on='anno.idx')
#     # REPLACE pack/unpack:
#     # df['anno.data'] = df['anno.data'].map(lambda x: json.dumps(x))
#     # REPLACE pack/unpack:
#     # df['anno.lbl.name'] = df['anno.lbl.name'].map(lambda x: json.dumps([x]))
    
#     if inplace:
#         self.df = df
#     return df


import os

import pandas as pd 
from joblib.parallel import Parallel, delayed, cpu_count
import numpy as np
from tqdm import tqdm
from lost_ds.functional.api import (remove_empty,
                               is_multilabel,
                               label_selection,
                               split_by_empty,
                               validate_empty_images
                               )
from lost_ds.im_util import get_imagesize, get_fs
from lost_ds.geometry.lost_geom import LOSTGeometries


def _get_and_validate_order(order, df, lbl_col, seg_lbl_col):
    
    if isinstance(order, list):
        order = {lbl: idx for idx, lbl in enumerate(order)}
    order = {k: v for k, v in sorted(order.items(), key=lambda item: item[1])}
    
    # if a multilabel column is passed we have to make sure that a maximum of 
    # one label is defined in the passed 'order'. If order definition is OK the 
    # multilabel column can be expressed by a single label column by only 
    # picking labels from order. Otherwise an Exception will be thrown
    
    if is_multilabel(df, lbl_col):
        # # remove dataset labels that do not occure in order
        df[seg_lbl_col] = df[lbl_col].map(lambda x: 
            [lbl for lbl in x if lbl in list(order.keys())])
        if df[seg_lbl_col].map(lambda x: len(x) > 1).sum():
            raise Exception('Found entries where multiple labels from order ' \
                'are defined! Pass an order without multilabel classes and run again')
        else:
            df[seg_lbl_col] = df[seg_lbl_col].map(lambda x: 
                x[0] if len(x) else np.nan)
    else:
        df[seg_lbl_col] = df[lbl_col]
        
    return order, df


def semantic_segmentation(order, dst_dir, fill_value, df, anno_dtypes=['polygon'], 
                          use_empty=False, lbl_col='anno_lbl', 
                          dst_path_col='seg_path', dst_lbl_col='seg_lbl', 
                          line_thickness=None, radius=None, filesystem=None, 
                          numeric_filename=False, parallel=-1):
    '''Create semantic segmentations from polygon-annos
    
    Args:
        df (pd.DataFrame): dataframe to generate the pixel maps from
        order (dict, list): order of the classes. Higher pixel
            values will overwrite lower ones. 
            Example: order=['Car', 'Person', 'Glasses'] or  
                order={'Car': 0, 'Person': 1, 'Glasses': 2}
                Car will get pixel value 0, Person px value 1 and so on - 
                Person overwrites Car, Glasses overwrites Person ... 
        dst_dir (str): Directory to store the pixel maps.
        fill_value (int): Pixel value for not annotated areas. Usually this
            will be something like 0 for something like 'background'.
        anno_dtypes (list of string): dtypes to use for segmentation. Possible
            values are {'polygon', 'bbox', 'line', 'point'}. Annotations with
            dtypes other than anno_dtypes will be removed from dataframe
        lbl_col (str): Column containing the training labels
        line_thickness (int): thickness of line-segmentation when using an 
            annotation of dtype line. Only takes effect when anno_dtypes 
            contains 'line'
        radius (int): radius of circle-segmentation when using an annotation
            of dtype point. Only takes effect when anno_dtypes contains 'point'
        filesystem (fsspec.filesystem, FileMan): filesystem to use. Use local
            if not initialized
        numeric_filename (bool): use same filename for seg and img if False. If
            True a numeric 12-digit filename will be generated. Usefull if some 
            files have same filenames.
        
    Returns:
        pd.DataFrame: The original dataframe with new column (dst_col) 
            containing the path to the according segmentation file. 
            Furthermore the column dst_lbl_col contains the label the
            segmentation looked up in order for creation
    '''
    fs = get_fs(filesystem)
    df_empty = None
    if use_empty:
        df = validate_empty_images(df=df)
        fill_value_key = list(order.keys())[list(order.values()).index(0)]
        df.loc[df[lbl_col].isnull(), dst_lbl_col] = [fill_value_key]
        df, df_empty = split_by_empty(df)        
    df = remove_empty(df=df, col='anno_data')
    order, df = _get_and_validate_order(order, df, lbl_col, dst_lbl_col)
    df = df[df.anno_dtype.isin(anno_dtypes)]
    
    # keep empty images if flag is set
    df = pd.concat([df, df_empty])
    
    def generate_seg(i, image_path, img_df):
        geom = LOSTGeometries()
        if not fs.exists(image_path):
            raise Exception('Image {} does not exist'.format(image_path))
        im_h, im_w = get_imagesize(image_path)
        segmentation = np.full([im_h, im_w], fill_value, dtype=np.int32)

        for label, level in order.items():
            draw = label_selection(labels=[label], df=img_df, col=dst_lbl_col)
            for _, row in draw.iterrows():
                segmentation = geom.segmentation(segmentation, level,
                                                 row.anno_data, row.anno_dtype, 
                                                 row.anno_format,row.anno_style,
                                                 line_thickness=line_thickness, 
                                                 radius=radius)
        if numeric_filename:
            filename = f'{i:012d}.png'
        else:
            filename = image_path.split('/')[-1].split('.')[0] + '.png'
        seg_path = os.path.join(dst_dir, filename)
        fs.write_img(segmentation, seg_path)
        return image_path, seg_path
    
    fs.makedirs(dst_dir, exist_ok=True)
    
    if parallel:
        paths = Parallel(n_jobs=parallel)(delayed(generate_seg)(i, path, img_df) 
                                        for i, (path, img_df) in enumerate(
                                            tqdm(
                                                df.groupby('img_path'), 
                                                desc='segmentation')
                                            )
                                        )
        path_map = {im_path: seg_path for im_path, seg_path in paths}
    else:
        path_map = dict()
        for i, (path, img_df) in enumerate(tqdm(df.groupby('img_path'), 
                                                desc='segmentation')
                                           ):
            p1, p2 = generate_seg(i, path, img_df)
            path_map[p1] = p2
    
    
    df[dst_path_col] = df.img_path.map(lambda x: path_map[x])
    return df



from lost_ds.vis.vis import (vis_and_store,
                             vis_sample,
                             vis_semantic_segmentation,
                             
                             )
from lost_ds.vis.geometries import (draw_boxes,
                                    draw_lines,
                                    draw_points,
                                    draw_polygons,
                                    draw_text)


import numpy as np

from shapely.geometry import box, Polygon

class Geometry(object):
    
    styles = ['xy']
    formats = ['abs', 'rel']
    
    def __init__(self):
        pass
        
        
    def _get_wh_struct(self, data, hw):
        h, w = hw[:2]
        assert isinstance(data, np.ndarray)
        data_len = data.shape[-1]
        assert data_len in [2, 4]
        wh_struct = [w, h] * (data_len // 2)
        return wh_struct


    def offset(self, data, offset):
        ''' Add an offset to the annotation
        
        Args:
            data (np.array): absolute data to transform
            offset (tuple): (H, W) offset to apply to the data
            
        Returns:
            np.ndarray with offset added
        '''
        wh_struct = self._get_wh_struct(data, offset)
        return data + wh_struct


    def to_abs(self, data, data_format, img_shape):
        ''' Transform relative data to absolute
        
        Args:
            data (np.array): relative data to transform
            data_format (str): format of data, one of {'abs', 'rel'}
            img_shape (tuple): (H,W) of image
            
        Returns:
            np.ndarray: data transformed to absolute
        '''
        new_data = None
        if data_format == 'abs':
            new_data = data
        else:
            wh_struct = self._get_wh_struct(data, img_shape)
            new_data = np.array(data * wh_struct)
        return new_data
    
    
    def to_rel(self, data, data_format, img_shape):
        ''' Transform absolute data to relative
        
        Args:
            data (np.ndarray): absolute data to transform
            data_format (str): format of data, one of {'abs', 'rel'}
            img_shape (tuple): (H,W) of image
            
        Returns:
            np.ndarray: data transformed to relative
        '''
        new_data = None
        if data_format == 'rel':
            new_data = data
        else:
            wh_struct = self._get_wh_struct(data, img_shape)
            new_data = np.array(data / wh_struct)
        return new_data
    
    
    def segmentation(self, *args, **kwargs):
        raise NotImplementedError('Method segmentation() at {}'.
                                  format(type(self)))
    
    
    def draw(self, img, data, data_format, style, text, color, line_thickness, 
             fontscale, **kwargs):
        data = self.to_abs(data, data_format, img.shape).astype(np.int32)
        img = self._draw(img, data, style, text, color, line_thickness, 
                         fontscale, **kwargs)
        assert img is not None
        return img
    
    
    def apply_crop(self, crop_pos:Polygon, data, style):
        return self.crop(crop_pos, data, style=style)


    def blow_up(self):
        pass
    
        
    def _crop_intersection(self, crop_pos, data, **kwargs):
        try:
            geo = self.to_shapely(data, **kwargs).buffer(0)
        except ValueError:
            return [np.nan]
        if geo.intersects(crop_pos):
            intersection = geo.intersection(crop_pos)
            return intersection, True
        else:
            return [np.nan], False
        



# def get_panoptic_data(self, is_thing, order=None, lbl_col='anno.lbl.name', 
#                         bbox_format='xywh', val_format='abs', df=None, 
#                         inplace=False):
#     '''Get panoptic data from df
#     Args:
#         is_thing (dict): mapping which categories can have instances 
#             e.g.:{'car': 1, 'person': 1, 'background': 0}
#         order (dict): A class to pixel mapping {class: px_val}, will be     
#             generated by unique labels in lbl_col if no order is provided
#             Note: You have to generate the pixelmap in the same manner
#         lbl_col (str): Column to look for the training labels
#         data_format (str): 'abs' - absolute panopt-data, 'rel' - relative
#         bbox_format (str): 'xywh' - topleftwh, 'lost' - midmidwh
#         df (pd.DataFrame): Frame to generate the pixel maps from
#         inplace (bool): Overwrite the own LOST-dataframe when True
#     Returns: 
#         tuple (pd.DataFrame, dict): Panoptic dataframe and order

#     '''
#     if df is None:
#         df = self.df 

#     if 'bbox' in df['anno.dtype'].unique():
#         df = df[df['anno.dtype'] == 'polygon']
#         print('WARNING: bboxes were found in dataframe. I removed them.')
#         assert(len(df) > 1), 'Dataframe has no polygon annotations!'
    
#     if order is None:
#         order = {l: i for i, l in enumerate(self.get_unique_labels(lbl_col))}

#     # unique image ids by img name
#     df['panopt.image_id'] = df['img.img_path'].map(lambda x: x.split('.')[0])

#     # polygon area per instance
#     df['panopt.area'] = df.map(lambda row: self._get_area(row), axis=1)
    
#     # bbox per instance
#     df['panopt.bbox'] = df['anno.data'].map(
#         lambda x: list(self._poly_to_bbox(x, bbox_format=bbox_format).values()))

#     # category_id, class id, we want to map it with the train-pixel value
#     df['panopt.category_id'] = df[lbl_col].map(lambda x: order[x])
    
#     # segment id per image in coco format 
#     df['panopt.id'] = None
#     for img_path, df_im in df.groupby('img.img_path'):
#         for category, df_cat in df_im.groupby('panopt.category_id'):
#             df_slice = df_im[df_im['panopt.category_id']==category]
#             df_slice['panopt.id'] = [i for i in range(len(df_slice))]
#             def get_segment_id(row):
#                 thing = is_thing[row[lbl_col]]
#                 return 1000 * row['panopt.category_id'] * thing + row['panopt.id']
#             df_slice['panopt.id'] = df_slice.apply(lambda row: get_segment_id(row), axis=1)
#             df_im[df_im['panopt.category_id']==category] = df_slice
#         df[df['img.img_path']==img_path] = df_im

#     # TODO: iscrowd
#     df['panopt.iscrowd'] = 0

#     if val_format == 'abs':
#         im_sizes = df['img.img_path'].map(lambda x: np.array(get_image_size(x)))

#         def _area_to_abs(area):
#             new_area = []
#             for i in range(len(area)):
#                 new_area.append(area[i] * im_sizes[i].prod())
#             return new_area

#         def _bbox_to_abs(bbox):
#             new_box = []
#             for i in range(len(bbox)):
#                 new_box.append(bbox[i] * np.hstack((im_sizes[i], im_sizes[i])) )
#             return new_box

#         df['panopt.bbox'] = _bbox_to_abs(df['panopt.bbox'])
#         df['panopt.area'] = _area_to_abs(df['panopt.area'])

#     if inplace:
#         self.df = df
#     return df, order

# def get_panoptic_segmentation(self, dst_dir, fill_value=0, df=None, inplace=False):
#     from panopticapi.utils import id2rgb
#     os.makedirs(dst_dir, exist_ok=True)
#     if df is None:
#         df = self.df
#     df = df[df['anno.dtype']=='polygon']
#     df['panopt.path'] = None
#     image_paths = self.get_unique_img_paths(df=df)
    
#     for i, image_path in enumerate(image_paths):    
#         print_progress_bar(i+1, len(image_paths), prefix='create panopt. maps:')
#         if not os.path.exists(image_path):
#             raise Exception('Image {} does not exist'.format(image_path))
#         im_w, im_h = get_image_size(image_path)
#         px_map = np.full([im_h, im_w], fill_value, dtype=np.int32)
#         image_annos = df[df['img.img_path']==image_path]

#         def draw_instance(instance):
#             poly = instance['anno.data']
#             poly = (np.array([list(point.values()) for point in poly]) * [im_w, im_h]).astype(np.int32)
#             cv2.fillPoly(px_map, [poly], instance['panopt.id'])

#         for i, instance in image_annos.iterrows():
#             draw_instance(instance)
#         px_map = id2rgb(px_map)
#         img_name = image_path.split('/')[-1]
#         filename = img_name.split('.')[0]
#         filename = filename + '.png'
#         mask_path = os.path.abspath(os.path.join(dst_dir, filename))
#         # cv2.imwrite(mask_path, px_map)
#         Image.fromarray(px_map).save(mask_path)
#         image_annos.loc[:, 'panopt.path'] = mask_path
#         df.loc[df['img.img_path']==image_path] = image_annos

#     if inplace:
#         self.df = df
#     return df


import os 

from lost_ds.functional.filter import (is_multilabel, 
                                      unique_labels,)


def remap_img_path(df, new_root_path, col='img_path', mode='replace'):
    """Remap image paths e.g. for environmental changes

    Args:
        df (pd.DataFrame): dataframe to apply remap
        new_root_path (string): path of the new dataset location
        col (str, optional): Column to remap. Defaults to 'img_path'.
        mode (str, optional): One of {'replace', 'prepend'}. 
            If 'replace' the path before the filename will be replaced by new_root_path.
            If 'prepend' new_root_path will be prepended to existing path.
            Defaults to 'replace'.

    Returns:
        pd.DataFrame: Dataframe with remapped path
    """
    df = df.copy()
    if mode == 'replace':
        df.loc[:, col] = df[col].map(lambda x: 
                os.path.join(new_root_path, os.path.basename(x)))
    elif mode == 'prepend':
        df.loc[:, col] = df[col].map(lambda x: os.path.join(new_root_path, x))
    else:
        raise Exception(NotImplementedError)
    return df
    
    
def remap_labels(df, label_mapping,  col='anno_lbl', dst_col='anno_lbl_mapped'):
    '''Map the labels in a given col
    Args:
        df (pd.DataFrame): dataframe to apply bbox transform
        label_mapping (dict): Direct mapping of labels {old: new}. Example:
            {'Cat':'Mammal', 'Dog':'Mammal', 'Car':'Vehicle', 'Tree': None}
            Note: labels with a mapping of None will be removed from dataset,
                    labels without any mapping wont be changed
        col (str): source column to read from
        dst_col (str): column to write to
        
    Returns:
        pd.DataFrame: The dataframe with the new mapped labels column lbl_col
    '''
    df = df.copy()
    label_mapping = label_mapping.copy()
    df[dst_col] = df[col]
    labels = unique_labels(df, col)
    # make sure every label occurs in the dictionary to prevent KeyError
    for l in labels:
        if not l in label_mapping.keys():
            label_mapping[l] = l
    if is_multilabel(df, dst_col):
        df[dst_col] = df[dst_col].map(lambda x: 
            [label_mapping[l] for l in x if not label_mapping[l] is None])
    else:
        df[dst_col] = df[dst_col].map(lambda x: label_mapping[x])
    return df



import numpy as np
import pandas as pd 

def is_multilabel(df, col):
    '''check if a column contains multilabels
    
    Args:
        df (pd.DataFrame): Dataframe to use for filtering
        col (str): column to check for multilabel
    
    Returns:
        bool: True if column contains multilabel, False if not
    '''
    if isinstance(df[col].iloc[0], (list, np.ndarray)):
        return True
    else:
        return False
    

def remove_empty(df, col='anno_data'):
    '''Remove images with empty columns in specified column
    
    Args:
        df (pd.DataFrame): Dataframe to use for filtering
        col (str): column to seatch for empty entries
    
    Returns:
        pd.DataFrame 
    '''
    return df[df[col].notnull()]


def unique_labels(df, col='anno_lbl'):
    '''Get unique dataset labels.
    
    Args:
        df (pd.DataFrame): dataframe to analyze
        col (str): Column containing list of labels

    Retrun:
        str: List of strings with unique classnames
    '''
    if is_multilabel(df=df, col=col):
        return np.unique(df[col].map(list).sum())
    return list(df[df[col].notna()][col].unique())
        
    
def selection_mask(labels, df, col='anno_lbl'):
    '''Get mask for labels in dataset
    Args:
        df (pd.DataFrame): dataset to mask
        col (str): Column containing list of labels
    
    Returns:
        pd.DataFrame: boolean mask defining which row contains one of the 
            provided labels
    '''
    if not isinstance(labels, (list, np.ndarray)):
        labels = [labels]
    if is_multilabel(df, col):
        # return df[col].map(lambda x: 
        #     bool(sum([l in list(x) for l in labels])))
        return df[col].map(lambda x: bool(len(np.intersect1d(x, labels))))
    else:
        return df[col].isin(labels)


def label_selection(labels, df, col='anno_lbl'):
    '''Get entries with a selection of labels from the dataframe
    Args:
        labels (list): list of labels to select
        df (pd.DataFrame): Frame to apply label selection
        col (str): Column containing list of labels
    Returns:
        pd.DataFrame: dataframe with label selection
    '''
    return df[selection_mask(df=df, labels=labels, col=col)]
    

def ignore_labels(labels, df, col='anno_lbl'):
    ''' Remove dataframe entries where the given labels occures

    Args:
        labels (list): list of labels to ignore
        df (pd.DataFrame): Frame to apply label ignore
        col (str): Column containing list of labels

    Returns:
        pd.DataFrame: dataframe with label selection
    '''
    return df[~selection_mask(df=df, labels=labels, col=col)]
    

def img_selection(imgs, df, inverse=False):
    '''Get entries with a selection of labels from the dataframe

    Args:
        imgs (list): list of imgs to select
        inverse (bool): get the selection if True, get the rest if False 
        df (pd.DataFrame): Frame to apply image selection

    Returns:
        pd.DataFrame: dataframe with image selection
    '''
    selection_mask = df.img_path.isin(imgs)
    if inverse:
        selection_mask = ~selection_mask
    return df[selection_mask]


import numpy as np

def bboxes_iou(boxes1, boxes2):
    boxes1 = np.array(boxes1)
    boxes2 = np.array(boxes2)

    boxes1_area = (boxes1[..., 2] - boxes1[..., 0]) * (boxes1[..., 3] - boxes1[..., 1])
    boxes2_area = (boxes2[..., 2] - boxes2[..., 0]) * (boxes2[..., 3] - boxes2[..., 1])

    left_up       = np.maximum(boxes1[..., :2], boxes2[..., :2])
    right_down    = np.minimum(boxes1[..., 2:], boxes2[..., 2:])

    inter_section = np.maximum(right_down - left_up, 0.0)
    inter_area    = inter_section[..., 0] * inter_section[..., 1]
    union_area    = boxes1_area + boxes2_area - inter_area
    ious          = np.maximum(1.0 * inter_area / (union_area + 1e-7), np.finfo(np.float32).eps)

    return ious


def bboxes_intersection(boxes1, boxes2):
    boxes1 = np.array(boxes1)
    boxes2 = np.array(boxes2)
    left_up       = np.maximum(boxes1[..., :2], boxes2[..., :2])
    right_down    = np.minimum(boxes1[..., 2:], boxes2[..., 2:])
    inter_section = np.maximum(right_down - left_up, 0.0)
    inter_area    = inter_section[..., 0] * inter_section[..., 1]
    return (inter_area > 0)


def _bbox_merge(bboxes):
    """Method to merge bboxes instead of nms them
    :param bboxes: (xmin, ymin, xmax, ymax, score, class)
    """
    bboxes = np.array(bboxes)
    
    if bboxes.size == 0:
        return np.array([])
    
    classes_in_img = list(set(bboxes[:, 5]))
    best_bboxes = []

    for cls in classes_in_img:
        cls_mask = (bboxes[:, 5] == cls)
        cls_bboxes = bboxes[cls_mask]
        bbox_mask = np.ones((len(cls_bboxes)))
        
        for box_idx in range(len(cls_bboxes)):
            if not bbox_mask[box_idx]:
                continue
            current_bbox = cls_bboxes[box_idx]
            other_ids = list(range(len(cls_bboxes)))
            other_ids.remove(box_idx)
            # bboxes[drop current box]
            other_bboxes = cls_bboxes[other_ids]
            # other_bboxes[drop alaready pruned boxes]
            # iou = bboxes_iou(current_bbox[None, :4], other_bboxes[list(bbox_mask[other_ids] != 0)][:, :4])
            intersection = bboxes_intersection(current_bbox[None, :4], other_bboxes[list(bbox_mask[other_ids] != 0)][:, :4])
            
            while any(intersection):
                valid_bboxes = bbox_mask[other_ids] != 0
                
                # get overlapping bboxes to combine
                gather_bboxes = other_bboxes[valid_bboxes][intersection]
                combine_bbox = np.concatenate((current_bbox[None], gather_bboxes), 0)
                xmin, xmax, ymin, ymax = combine_bbox[:, [0, 2]].min(), combine_bbox[:, [0, 2]].max(), combine_bbox[:, [1, 3]].min(), combine_bbox[:, [1, 3]].max()
                score = combine_bbox[:, 4].max()
                current_bbox = np.array([xmin, ymin, xmax, ymax, score, cls])
                cls_bboxes[box_idx] = current_bbox
                
                # remember pruned bboxes
                combined_ids = np.array(other_ids)[valid_bboxes][intersection]
                bbox_mask[combined_ids] = 0
                
                # find iou for next iteration
                remaining_bboxes = other_bboxes[valid_bboxes][~intersection] # iou < iou_threshold]
                # iou = bboxes_iou(current_bbox[None, :4], remaining_bboxes[:, :4])
                intersection = bboxes_intersection(current_bbox[None, :4], remaining_bboxes[:, :4])
        
        best_bboxes.append(cls_bboxes[bbox_mask != 0])

    return np.array(best_bboxes).reshape((-1, 6))


def bbox_merge(boxes_list, scores_list, labels_list, iou_thr, **kwargs):
    for bboxes, scores, labels in zip(boxes_list, scores_list, labels_list):
        bboxes = np.array(bboxes).reshape((-1, 4))
        scores = np.array(scores_list).reshape((-1, 1))
        labels = np.array(labels_list).reshape((-1, 1))
        merged = _bbox_merge(np.hstack((bboxes, scores, labels)))
        return merged[:, :4], merged[:, 4], merged[:, 5]


from lost_ds.geometry.geometry import Geometry
from lost_ds.geometry.bbox import Bbox
from lost_ds.geometry.polygon import Polygon
from lost_ds.geometry.point import Point
from lost_ds.geometry.line import Line
from lost_ds.geometry.lost_geom import LOSTGeometries


import multiprocessing as mp 

from dask.distributed import Client, LocalCluster

def get_dask_client(client, cluster):
    if client is None:
        if cluster is None:
            cluster = LocalCluster(processes=False, asynchronous=True, 
                                   n_workers=mp.cpu_count())
        client = Client(cluster)
    elif 'Client' in str(client):
        client = client
    else:
        raise ValueError('Client {} is not supported!'.format(client))
    return client


import os 
import json
from multiprocessing import Process
from threading import Thread
from typing import Union

from tqdm import tqdm
import pandas as pd 
import numpy as np
import fsspec
from joblib import Parallel, delayed, cpu_count
from lost_ds.copy import copy_imgs
from lost_ds.functional.validation import validate_img_paths, validate_single_labels, validate_unique_annos
from lost_ds.functional.filter import remove_empty

from lost_ds.geometry.api import LOSTGeometries
from lost_ds.geometry.bbox import Bbox
from lost_ds.im_util import get_imagesize
from lost_ds.io.file_man import FileMan
from lost_ds.functional.filter import is_multilabel, unique_labels
from lost_ds.util import get_fs

        
# def to_abs(df, path_col='img_path', 
#            filesystem:Union[FileMan,fsspec.AbstractFileSystem]=None, 
#            verbose=True, parallel=-1):
#     ''' Transform all annos to absolute annos
    
#     Args:
#         df (pd.DataFrame): dataframe to transform
#         filesystem (fsspec.filesystem, FileMan): filesystem to use. Use local
#             if not initialized
#         verbose (bool): print tqdm progress-bar
            
#     Returns:
#         pd.DataFrame: transformed dataframe wil absolute annotations
#     '''
#     df = df.copy()
#     df_rel = df[df['anno_format'] == 'rel']
#     cols = [path_col, 'anno_data', 'anno_dtype', 'anno_format']
#     def make_abs(row):
#         anno_data = row.anno_data
#         anno_format = row.anno_format
#         if isinstance(anno_data, np.ndarray) and anno_format=='rel':
#             geom = LOSTGeometries()
#             img_path = row[path_col]
#             anno_dtype = row.anno_dtype
#             img_shape = get_imagesize(img_path, filesystem)
#             anno_data = geom.to_abs(
#                 anno_data, anno_dtype, anno_format, img_shape
#                 )    
#         return anno_data
    
#     ### joblib workflow
#     if parallel:
#         abs_data = Parallel(parallel)(delayed(make_abs)(row)
#                                 for idx, row in tqdm(df_rel[cols].iterrows(), 
#                                                     total=len(df_rel),
#                                                     desc='to abs', 
#                                                     disable=(not verbose)))
#     else:
#         abs_data = list()
#         for idx, row in tqdm(df_rel[cols].iterrows(), total=len(df_rel),
#                             desc='to abs', disable=(not verbose)):
#             abs_data.append(make_abs(row))
            
#     df.loc[df['anno_format'] == 'rel', 'anno_data'] = pd.Series(abs_data, 
#                                                                 index=df_rel.index,
#                                                                 dtype=object)
#     df.loc[df['anno_format'] == 'rel', 'anno_format'] = 'abs'
#     ###

#     ### apply workflow
#     # df.loc[df['anno_format'] == 'rel', ['anno_data', 'anno_format']] = df.loc[df['anno_format'] == 'rel', cols].map(lambda x: make_abs(x, True), axis=1, result_type='expand')
#     ###
    
#     return df

def to_abs(df, path_col='img_path', 
           filesystem:Union[FileMan,fsspec.AbstractFileSystem]=None, 
           verbose=True, parallel=-1):
    ''' Transform all annos to absolute annos
    
    Args:
        df (pd.DataFrame): dataframe to transform
        filesystem (fsspec.filesystem, FileMan): filesystem to use. Use local
            if not initialized
        verbose (bool): print tqdm progress-bar
            
    Returns:
        pd.DataFrame: transformed dataframe wil absolute annotations
    '''
    df = df.copy()
    df_rel = df[df['anno_format'] == 'rel']
    
    def make_abs(path, df):
        df = df.copy()
        h,w = get_imagesize(path, filesystem)
        xy = np.array([w,h])
        def _abs(data):
            if isinstance(data, np.ndarray):
                sh = data.shape
                return (data.reshape((-1,2)) * xy).reshape(sh)
            else: 
                return data
        df['anno_data'] = df['anno_data'].map(_abs)
        df['anno_format'] = 'abs'
        return df
    
    if parallel:
        abs_dfs = Parallel(parallel)(delayed(make_abs)(path, path_df) 
                                     for path, path_df in tqdm(df_rel.groupby(path_col), 
                                                               total=len(df_rel[path_col].unique()),
                                                               desc='to abs',
                                                               disable=(not verbose)))
    else:
        abs_dfs = list()
        for path, path_df in tqdm(df_rel.groupby(path_col), 
                                  total=len(df_rel[path_col].unique()),
                                  desc='to abs', 
                                  disable=(not verbose)):
            abs_dfs.append(make_abs(path, path_df))
    
    abs_dfs.append(df[df['anno_format']!='rel'])
    df_abs = pd.concat(abs_dfs)
    return df_abs


# def to_rel(df, path_col='img_path', 
#            filesystem: Union[FileMan, fsspec.AbstractFileSystem] = None,
#            verbose=True, parallel=-1):
#     ''' Transform all annos to absolute annos
    
#     Args:
#         df (pd.DataFrame): dataframe to transform
#         filesystem (FileMan): Filesystem instance. If None local filesystem 
#             is used
#         verbose (bool): print tqdm progress-bar
            
#     Returns:
#         pd.DataFrame: transformed dataframe wil absolute annotations
#     '''
#     df = df.copy()
#     df_abs = df[df['anno_format'] == 'abs']
#     cols = [path_col, 'anno_data', 'anno_dtype', 'anno_format']
#     def make_rel(row):
#         anno_data = row.anno_data
#         anno_format = row.anno_format
#         if isinstance(anno_data, np.ndarray) and anno_format == 'abs':
#             geom = LOSTGeometries()
#             img_path = row[path_col]
#             anno_dtype = row.anno_dtype
#             img_shape = get_imagesize(img_path, filesystem)
#             anno_data = geom.to_rel(
#                 anno_data, anno_dtype, anno_format, img_shape
#             )
#         return anno_data
    
#     if parallel:
#         rel_data = Parallel(parallel)(delayed(make_rel)(row) 
#                                 for idx, row in tqdm(df_abs[cols].iterrows(), 
#                                                     total=len(df_abs),
#                                                     desc='to rel',
#                                                     disable=(not verbose)))
#     else:
#         rel_data = list()
#         for idx, row in tqdm(df_abs[cols].iterrows(), total=len(df_abs),
#                              desc='to rel',disable=(not verbose)):
#             rel_data.append(make_rel(row))
            
#     df.loc[df['anno_format'] == 'abs', 'anno_data'] = pd.Series(rel_data, 
#                                                                 index=df_abs.index,
#                                                                 dtype=object)
#     df.loc[df['anno_format'] == 'abs', 'anno_format'] = 'rel'
    
#     return df


def to_rel(df, path_col='img_path', 
           filesystem: Union[FileMan, fsspec.AbstractFileSystem] = None,
           verbose=True, parallel=-1):
    ''' Transform all annos to absolute annos
    
    Args:
        df (pd.DataFrame): dataframe to transform
        filesystem (FileMan): Filesystem instance. If None local filesystem 
            is used
        verbose (bool): print tqdm progress-bar
            
    Returns:
        pd.DataFrame: transformed dataframe wil absolute annotations
    '''
    df = df.copy()
    df_abs = df[df['anno_format'] == 'abs']
    
    def make_rel(path, df):
        df = df.copy()
        h,w = get_imagesize(path, filesystem)
        xy = np.array([w,h])
        def _rel(data):
            if isinstance(data, np.ndarray):
                sh = data.shape
                return (data.reshape((-1,2)) / xy).reshape(sh)
            else: 
                return data
        df['anno_data'] = df['anno_data'].map(_rel)
        df['anno_format'] = 'rel'
        return df
    
    if parallel:
        rel_dfs = Parallel(parallel)(delayed(make_rel)(path, path_df) 
                                     for path, path_df in tqdm(df_abs.groupby(path_col), 
                                                               total=len(df_abs[path_col].unique()),
                                                               desc='to rel',
                                                               disable=(not verbose)))
    else:
        rel_dfs = list()
        for path, path_df in tqdm(df_abs.groupby(path_col), 
                                  total=len(df_abs[path_col].unique()),
                                  desc='to rel', 
                                  disable=(not verbose)):
            rel_dfs.append(make_rel(path, path_df))
    
    rel_dfs.append(df[df['anno_format']!='abs'])
    df_rel = pd.concat(rel_dfs)
    return df_rel


def transform_bbox_style(dst_style, df):
    ''' Transform all bbox annos to a specified style
    
    Args:
        dst_style (str): desired bbox style: {'xywh', 'x1y1x2y2', 'xcycwh'}
        df (pd.DataFrame): Dataframe to transform bbox styles
            
    Returns:
        pd.DataFrame: transformed dataframe with bboxes in new format
    '''
    df = df.copy()
    geom = LOSTGeometries()
    cols = ['anno_data', 'anno_style']
    df_bbox = df.loc[df.anno_dtype == 'bbox', cols]
    new_data = []
    new_style = []
    for idx, row in df_bbox.iterrows():
        if isinstance(row.anno_data, np.ndarray):
            data = geom.transform_bbox_style(row.anno_data, 
                                            row.anno_style, 
                                            dst_style)
            style = dst_style
        else:
            data = row.anno_data
            style = row.anno_style
        new_data.append(data)
        new_style.append(style)
        
    df_bbox['anno_data'] = new_data
    df_bbox['anno_style'] = new_style
    df.loc[df.anno_dtype=='bbox', cols] = df_bbox[cols]
    return df


def polygon_to_bbox(df, dst_style=None):
    '''Transform all polygons to bboxes
    Args:
        df (pd.DataFrame): datafram to apply transform
        lost_geometries (LOSTGeometries, None): Use an existing instance of
            LOSTGeometries if passed. Otherwise creates a new one during RT.
        dst_style (str): desired style of bbox: {'xywh', 'x1y1x2y2', 'xcycwh'}
    Returns:
        pd.DataFrame with transformed polygons
    '''
    df = df.copy()
    geom = LOSTGeometries()
    if dst_style is None:
        dst_style = 'x1y1x2y2'
    polygons = df[df.anno_dtype == 'polygon']
    df.loc[df.anno_dtype=='polygon', 'anno_data'] = polygons.anno_data.map(
        lambda x: geom.poly_to_bbox(x, dst_style))
    # df.loc[df.anno_dtype=='polygon', 'anno_data'] = polygons.anno_data.map(
    #     geom.poly_to_bbox, dst_style)
    columns = ['anno_style', 'anno_dtype']
    df.loc[df.anno_dtype=='polygon', columns] = [dst_style, 'bbox']
    return df


# TODO: allow iscrowd 
# TODO: implement segmentation to RLE for iscrowd==1 and results format
# TODO: enable panoptic segmentation
def to_coco(df, remove_invalid=True, lbl_col='anno_lbl', 
            predef_img_mapping:dict=None,
            predef_lbl_mapping:dict=None,
            supercategory_mapping=None, 
            copy_path=None, 
            copy_images=False, 
            link_images=True, 
            rename_by_index=True,
            json_path=None,
            filesystem: Union[FileMan, fsspec.AbstractFileSystem] = None):
    """Transform dataset to coco data format

    Args:
        df ([pd.DataFrame]): dataframe to apply transform
        remove_invalid (bool, optional): Remove invalid image-paths. Defaults to True.
        predef_img_mapping (dict, optional): predefined uid mapping for images like {img_path: uid}
        predef_lbl_mapping (dict, optional): predefined uid mapping for labels like {lbl: uid}
        lbl_col (str, optional): dataframe column to look for labels. Defaults to 'anno_lbl'.
        supercategory_mapping ([dict], optional): dict like mapping for supercategories ({class: superclass}). 
            Defaults to None.
        copy_path (str, optional): output directory to copy or link images
        copy_images (bool, optional): copy images if True
        link_images (bool, optional): symlink images instead of copy if True
        rename_by_index (bool, optional): rename files by index with pattern 
            {img_id:012d}-{file_base}.{file_ending} to assure unique filenames
        json_path (str, optional): store coco annotations as .json file 
        filesystem (Union[FileMan, fsspec.AbstractFileSystem], optional): Filesystem instance. 
            Defaults to None.

    Returns:
        dict: containing coco data like {'categories': [...], 'images': [...], 'annotations': [...]}
    """
    filesystem = get_fs(filesystem)
    df = validate_single_labels(df, lbl_col, 'coco_lbl')
    df = validate_unique_annos(df)
    lbl_col = 'coco_lbl'
    assert not is_multilabel(df, lbl_col), 'Provided lbl-col {} contains multilabels'
    df = validate_img_paths(df, remove_invalid, filesystem)
    df = transform_bbox_style('xywh', df)
    df = to_abs(df, filesystem=filesystem, verbose=False)
    
    if predef_img_mapping is None:
        img_mapping = {img: idx for idx, img in enumerate(df['img_path'].unique())}
    else:
        img_mapping = predef_img_mapping
        
    categories = list()
    imgs = list()
    annos = list()
    
    # COCO-categories
    if predef_lbl_mapping is None:
        u_lbls = unique_labels(df[(df['anno_data'].notnull()) & (df[lbl_col].notnull())], lbl_col)
        lbl_to_id = {lbl: lbl_id+1 for lbl_id, lbl in enumerate(u_lbls)}
    else:
        lbl_to_id = predef_lbl_mapping
    for lbl, lbl_id in lbl_to_id.items():
        supercategory = None if supercategory_mapping is None else supercategory_mapping[lbl]
        categories.append({"id": lbl_id, 
                           "name": lbl,
                           "supercategory": supercategory})
    
    copy_process = list()
    if copy_path is None:
        root_dirs = df['img_path'].map(lambda x: '/'.join(x.split('/')[:-1])).unique()
        # assert len(root_dirs) == 1, f'COCO-Dataset images located in multiple different dirs {root_dirs}. ' \
        #     'You can use copy_path argument to copy the entire dataset.'
    else:
        filesystem.makedirs(copy_path, True)
        
    for img_path, img_id in img_mapping.items():
                
        # COCO-imgs
        im_h, im_w = get_imagesize(img_path)
        filename = img_path.split('/')[-1]
        if copy_path is not None:
            if rename_by_index:
                file_ending = img_path.split('.')[-1]
                file_base = filename.split('.')[0]
                filename = f'{img_id:012d}-{file_base}.{file_ending}'
            dst_file = os.path.join(copy_path, filename)
            # if not filesystem.exists(dst_file):
            # p = Process(target=filesystem.copy, args=(img_path, dst_file,), daemon=True)
            if copy_images:
                p = Thread(target=filesystem.copy, args=(img_path, dst_file,), daemon=True)
                copy_process.append(p)
                p.start()
            elif link_images: 
                os.symlink(img_path, dst_file)
            
        imgs.append({"org_path": img_path,      # custom
                     "id": img_id,              # int, 
                     "width": im_w,             # int, 
                     "height": im_h,            # int, 
                     "file_name": filename,     # str, 
                     "license": None,           # int, 
                     "flickr_url": None,        # str, 
                     "coco_url": None,          # str, 
                     "date_captured": None,     # datetime,
                     })
        
        # COCO-annos
        img_df = df[df['img_path']==img_path].copy()
        if not img_df['anno_data'].notnull().any():
            continue
        img_df = remove_empty(img_df)
        for _, row in img_df.iterrows():
            anno_type = row['anno_dtype']
            anno = {"id": len(annos) + 1,
                    "image_id": img_id,
                    "category_id": lbl_to_id[row[lbl_col]],
                    "iscrowd": 0,
                    }
            area = 0
            if 'anno_confidence' in row.keys():
                if row['anno_confidence']:
                    if row['anno_confidence'] > 0:
                        score = float(row['anno_confidence'])
                        anno['score'] = score
            if anno_type == 'polygon':
                segmentation = [row['anno_data'].flatten().tolist()]
                bbox = list(LOSTGeometries().poly_to_bbox(row['anno_data'], 'xywh'))
                area = bbox[2] * bbox[3]
                anno['segmentation'] = segmentation
                anno['bbox'] = bbox
                anno['area'] = area
            elif anno_type == 'bbox':
                bbox = list(LOSTGeometries().transform_bbox_style(row['anno_data'], row['anno_style'], 'xywh'))
                area = bbox[2] * bbox[3]
                anno['bbox'] = bbox
                anno['area'] = area
            elif anno_type is None:
                pass
            else:
                raise Exception(f'Unsupported type {anno_type}')
            if area:
                annos.append(anno)

    coco_anno = {"info": {"year": None, 
                          "version": None, 
                          "description": None, 
                          "contributor": None, 
                          "url": None, 
                          "date_created": None
                          },
                 "licenses": [{"id": None, 
                               "name": None, 
                               "url": None
                               }],
                 "categories": categories,
                 "images": imgs,
                 "annotations": annos,
                 }
    
    if copy_path:
        [p.join() for p in copy_process]
    
    if json_path:
        anno_dir = os.path.dirname(json_path)
        if not filesystem.exists(anno_dir):
            filesystem.makedirs(anno_dir, True)
        with open(json_path, 'w') as f:
            json.dump(coco_anno, f) 
    
    return coco_anno


import numpy as np
import pandas as pd 

from lost_ds.io.file_man import FileMan
from lost_ds.geometry.api import LOSTGeometries

from lost_ds.functional.validation import (validate_geometries,
                                           validate_empty_images,
                                           validate_unique_annos,
                                           validate_img_paths,
                                           validate_single_labels)

from lost_ds.functional.split import (split_by_empty,
                                      split_by_img_path,
                                      split_train_test,
                                      split_train_test_multilabel,
                                      split_multilabels)

from lost_ds.functional.filter import (remove_empty,
                                       ignore_labels,
                                       img_selection,
                                       is_multilabel,
                                       label_selection,
                                       selection_mask,
                                       unique_labels)

from lost_ds.functional.transform import (polygon_to_bbox,
                                          to_abs, to_coco,
                                          to_rel,
                                          transform_bbox_style)

from lost_ds.functional.mapping import (remap_img_path,
                                       remap_labels,
                                       )

from lost_ds.copy import (copy_imgs,
                          pack_ds
                          )

from lost_ds.vis.vis import (vis_and_store,
                             vis_semantic_segmentation,
                             )

from lost_ds.cropping.cropping import (crop_dataset,
                                       DSCropper,
                                       crop_components,
                                       )

from lost_ds.segmentation.semantic_seg import (semantic_segmentation,
                                               )

from lost_ds.segmentation.anno_from_seg import (segmentation_to_lost,
                                               )

from lost_ds.detection.detection import detection_dataset, bbox_nms

from lost_ds.masking.masking import mask_dataset

from lost_ds.util import to_parquet


class LOSTDataset(object):
    '''Represents a dataset in LOST format.'''
    
    def __init__(self, 
                 ds, 
                 filesystem=None, 
                 **kwargs):
        ''' Initialize LOSTDataset
        
        Args:
            ds (str, pd.DataFrame, LOSTDataset, list(s)): 
                Dataset(s) to use. The different types can be combined in lists.
            filesystem (fsspec.filesystem, None): Filesystem to use. If None the 
                local filesystem will be used.
            
        '''
        self.fileman = FileMan(filesystem)
        self.df = self.fileman.load_dataset(ds)
        try:
            self._parse_data()
        except TypeError as e:
            print('parsing: ', e)
        self.geom = LOSTGeometries()
        self.cropper = DSCropper(self.geom, self.fileman)
        
    #
    #   self.df management
    #
    
    def _parse_data(self):
        # parse lds-serialized data (2D-lists) to numpy array
        def _parse(data):
            try: 
                return np.vstack(data).squeeze()
            except:
                return data
        
        parse_keys = [k for k in self.df.keys() if k.endswith('_lds_serialized')]
        if parse_keys:
            col_mapper = dict()
            for k in parse_keys:
                self.df[k] = self.df[k].map(_parse)
                col_mapper[k] = k.replace('_lds_serialized', '')
            self.df.rename(col_mapper, axis=1, inplace=True)
            
        if 'anno_data' in self.df:
            self.df['anno_data'] = self.df['anno_data'].map(_parse)
        
            
    def to_parquet(self, path, df=None):
        ''' Store dataset as .parquet file
        
        Args:
            path (str): path to store the dataset
            df (pd.DataFrame): dataframe to store
        '''
        df = self._get_df(df)
        to_parquet(path, df, self.fileman)
        
    
    def _get_df(self, df):
        '''DataFrame selection
        
        Args:
            df (pd.DataFrame, None): dataframe to select
            
        Returns:
            pd.DataFrame: self.df if df is None else df
        '''
        if df is None:
            return self.df.copy()
        elif isinstance(df, LOSTDataset):
            return df.df
        elif isinstance(df, pd.DataFrame):
            return df
    
    
    def _update_inplace(self, df, inplace):
        '''inplace management
        
        Args:
            df (pd.DataFrame): dataframe to update
            inplace (bool): Flag if self.df should be overwritten
            
        Returns:
            None if inplace=True
            pd.DataFrame if inplace=False
        '''
        if inplace:
            self.df = df
            return None
        else:
            return df
    
    
    def get_imagesize(self, path):
        '''Get the size of an image without reading the entire data
        
        Args:
            path (str): path to image
            
        Returns:
            tuple of int: (im_h, im_w) height and width of the image
        '''
        return self.fileman.get_imagesize(path)
        
    
    def copy(self):
        return LOSTDataset(self.df.copy(), self.fileman)
    
    
    #
    #   adapt important and usefull pandas- and builtin-functions
    #
    
    def __repr__(self):
        # enables printing dataframe in interpreter by 'ds'
        return self.df.__repr__()
    
    def __str__(self):
        # enabels printing dataframe by 'print(ds)'
        return str(self.df)
    
    def __len__(self):
        # enables getting dataframe length
        return len(self.df)

    def keys(self):
        return self.df.keys()
    
    def __getitem__(self, name):
        # enables getting value by 'ds[key]'
        return self.df[name]
    
    def __setitem__(self, key, value):
        return self.df.__setitem__(key, value)
    
    # def __getattr__(self, name: str):
    #     # enable getting attributes with ds.attr
    #     # if name.startswith('__') and name.endswith('__'):
    #     #     raise ValueError
    #     try:
    #         return self.df.__getattr__(name)
    #     except:
    #         try:
    #             return getattr(self, name)
    #         except:
    #             raise AttributeError(name)
    
    #
    #   Data copy
    #
    
    def copy_imgs(self, out_dir, df=None, col='img_path', 
                  force_overwrite=False, parallel=-1):
        '''Copy all images of dataset into out_dir

        Args:
            df (pd.DataFrame): dataframe to copy
            out_dir (str): Destination folder to store images
            col (str): column containing paths to files
        '''
        df = self._get_df(df)
        copy_imgs(df=df, out_dir=out_dir, col=col, 
                  force_overwrite=force_overwrite, filesystem=self.fileman, 
                  parallel=parallel)
            
    
    def pack_ds(self, out_dir, df=None, cols=['img_path', 'mask_path'], 
                dirs = ['imgs', 'masks', 'crops'], inplace=False):
        '''Copy all images from dataset to a new place and update the dataframe 
        
        Args:
            df (pd.DataFrame): Dataframe to copy
            out_dir (str): Name of the directory to store the information
            cols (list of string): column names containing file-paths
            dirs (list of string): name of new directories according to cols. 
                The dirs will contain the copied data
        
        Returns:
            pd.DataFrame with new image paths
        '''
        df = self._get_df(df)
        df = pack_ds(df=df, out_dir=out_dir, cols=cols, dirs=dirs, 
                     filesystem=self.fileman)
        return self._update_inplace(df, inplace=inplace)
    
    
    #
    #   Filtering
    #
    
    def remove_empty(self, df=None, col='anno_data', inplace=False):
        '''Remove images with empty columns in specified column
        
        Args:
            df (pd.DataFrame): Dataframe to use for filtering
            col (str): column to seatch for empty entries
            inplace (bool): overwrite self.df with result if True
        
        Returns:
            pd.DataFrame if inplace=False
            None if inplace=True
        '''
        df = self._get_df(df)
        df = remove_empty(df, col)
        return self._update_inplace(df, inplace)
    
    
    def ignore_labels(self, labels, df=None, col='anno_lbl', inplace=False):
        ''' Remove dataframe entries where the given labels occures

        Args:
            labels (list): list of labels to ignore
            df (pd.DataFrame): Frame to apply label ignore
            col (str): Column containing list of labels
            inplace (bool): overwrite self.df with result if True

        Returns:
            pd.DataFrame: dataframe with label selection
        '''
        df = self._get_df(df)
        df = ignore_labels(labels=labels, df=df, col=col)
        return self._update_inplace(df, inplace)
    
    
    def selection_mask(self, labels, df=None, col='anno_lbl'):
        '''Get mask for labels in dataset
        Args:
            df (pd.DataFrame): dataset to mask
            col (str): Column containing list of labels
        
        Returns:
            pd.DataFrame: row-wise boolean mask with True is a label occures in
                this row, False if not
        '''
        df = self._get_df(df)
        return selection_mask(labels=labels, df=df, col=col)
    
    
    def img_selection(self, imgs, df=None, inverse=False, inplace=False):
        '''Get entries with a selection of labels from the dataframe

        Args:
            imgs (list): list of imgs to select
            inverse (bool): get the selection if True, get the rest if False 
            df (pd.DataFrame): Frame to apply image selection

        Returns:
            pd.DataFrame: dataframe with image selection
        '''
        df = self._get_df(df)
        df = img_selection(imgs=imgs, df=df, inverse=inverse)
        return self._update_inplace(df, inplace)
    
    
    def is_multilabel(self, df=None, col='anno_lbl'):
        '''check if a column contains multilabels

        Args:
            df (pd.DataFrame): Dataframe to use for filtering
            col (str): column to check for multilabel

        Returns:
            pd.DataFrame: row-wise boolean mask with True if column contains 
                multilabel False if not
        '''
        df = self._get_df(df)
        return is_multilabel(df=df, col=col)
        
    
    def label_selection(self, labels, df=None, col='anno_lbl', inplace=False):
        '''Get entries with a selection of labels from the dataframe
        Args:
            labels (list): list of labels to select
            df (pd.DataFrame): Frame to apply label selection
            col (str): Column containing list of labels
        Returns:
            pd.DataFrame: dataframe with label selection
        '''
        df = self._get_df(df)
        df = label_selection(labels=labels, df=df, col=col)
        return self._update_inplace(df, inplace)
    
    
    def unique_labels(self, df=None, col='anno_lbl'):
        '''Get unique dataset labels.
        Args:
            df (pd.DataFrame): dataframe to analyze
            col (str): Column containing list of labels

        Retrun:
            str: List of strings with unique classnames
        '''
        df = self._get_df(df)
        return unique_labels(df, col)
        
    
    
    #
    #   Mapping
    #
    
    def remap_img_path(self, new_root_path, df=None, col='img_path', 
                       inplace=False):
        '''Remap img_path by prepending a root path to the filename
        
        Args:
            new_root_path (str): New root path for all files 
            col (str): The column that should be remaped
        '''
        df = self._get_df(df)
        df = remap_img_path(df, new_root_path, col)
        return self._update_inplace(df, inplace)
    
    def remap_labels(self, label_mapping, df=None, col='anno_lbl', 
                     dst_col='anno_lbl_mapped', inplace=False):
        '''Map the labels in a given col
        
        Args:
            df (pd.DataFrame): dataframe to apply bbox transform
            label_mapping (dict): Direct mapping of labels {old: new}. Example:
                {'Cat':'Mammal', 'Dog':'Mammal', 'Car':'Vehicle', 'Tree': None}
                Note: labels with a mapping of None will be removed from dataset,
                        labels without any mapping wont be changed
            col (str): source column to read from
            dst_col (str): column to write to
            
        Returns:
            pd.DataFrame: The dataframe with the new mapped labels column lbl_col
        '''
        df = self._get_df(df)
        df = remap_labels(df, label_mapping, col, dst_col)
        return self._update_inplace(df, inplace)
    
    
    #
    #   Transformation
    #
    
    def to_abs(self, df=None, path_col='img_path', verbose=True, inplace=False,
               parallel=-1):
        ''' Transform dataframe to absolute data
        
        Args:
            df (pd.DataFrame): dataframe to transform
            verbose (bool): print tqdm progress-bar
            inplace (bool): overwrite self.df
        '''
        df = self._get_df(df)
        df = to_abs(df, path_col, self.fileman, verbose, parallel=parallel)
        return self._update_inplace(df, inplace)
    
    
    def to_rel(self, df=None, path_col='img_path', verbose=True, inplace=False,
               parallel=-1):
        ''' Transform dataframe to relative data
        
        Args:
            df (pd.DataFrame): dataframe to transform
            verbose (bool): print tqdm progress-bar
            inplace (bool): overwrite self.df
        '''
        df = self._get_df(df)
        df = to_rel(df, path_col, self.fileman, verbose, parallel=parallel)
        return self._update_inplace(df, inplace)
    
    
    def polygon_to_bbox(self, df=None, dst_style=None, inplace=False):
        '''Transform all polygons to bboxes
        
        Args:
            df (pd.DataFrame): datafram to apply transform
            dst_style (str): desired style of bbox: 
                {'xywh', 'x1y1x2y2', 'xcycwh'}
            inplace (bool): overwrite self.df
            
        Returns:
            pd.DataFrame with transformed polygons
        '''
        df = self._get_df(df)
        df = polygon_to_bbox(df, dst_style)
        return self._update_inplace(df, inplace)
    
    
    def transform_bbox_style(self, dst_style, df=None, inplace=False):
        '''Transform anno-style of all bboxes
        
        Args:
            dst_style (str): desired bbox style {'xcycwh', 'xywh', 'x1y1x2y2'}
            df (pd.DataFrame): dataframe to transform
            inplace (bool): overwrite self.df
        
        Returns:
            pd.DataFrame: dataframe with transformed bboxes
        '''
        df = self._get_df(df)
        df = transform_bbox_style(dst_style, df)
        return self._update_inplace(df, inplace)
        
        
    def to_coco(self, df=None, remove_invalid=True, lbl_col='anno_lbl', 
                supercategory_mapping=None, copy_path=None, json_path=None,
                rename_by_index=False):
        """Transform dataset to coco data format

        Args:
            df ([pd.DataFrame], optional): dataframe to apply transform
            remove_invalid (bool, optional): Remove invalid image-paths. Defaults to True.
            lbl_col (str, optional): dataframe column to look for labels. Defaults to 'anno_lbl'.
            supercategory_mapping ([type], optional): dict like mapping for supercategories ({class: superclass}). 
                Defaults to None.
            copy_path (str, optional): copy all images to given directory
            rename_by_index (bool, optional): rename files by index to assure unique filenames
            json_path (str, optional): store coco annotations as .json file 

        Returns:
            dict: dict containing coco data like {'categories': [...], 'images': [...], 'annotations': [...]}
        """
        df = self._get_df(df)
        coco_annos = to_coco(df, remove_invalid=remove_invalid, lbl_col=lbl_col, 
                             supercategory_mapping=supercategory_mapping, 
                             copy_path=copy_path, 
                             rename_by_index=rename_by_index, 
                             json_path=json_path,
                             filesystem=self.fileman)
        return coco_annos
    
    
    #
    #   Splitting
    #
    
    def split_by_empty(self, df=None, col='anno_data'):
        '''Split dataframe into empty and non-empty entries by given collumn
        
        Args:
            df (pd.DataFrame): df to apply on
            col (str): column to check for empty entries
            
        Returns:
            tuple of pd.DataFrame: first dataframe contains non-emtpy entries, 
                the second dataframe contains empty entries
        '''
        df = self._get_df(df)
        return split_by_empty(df, col)
    
    def split_train_test(self, test_size=0.2, val_size=0.2, stratify_col=None, 
                         df=None, col='img_path', random_state=42):
        '''Split dataset based on img paths (for dataset with multiple 
            entries for one image)
        Args:
            test_size (float): fraction of images in df that will be used for 
                test dataset
            val_size (float): fraction of images in df that will be used for 
                test dataset
            stratify_col (str): column to use for stratify split
            df (pd.DataFrame): Dataframe to split
        Returns: 
            tuple: pd.DataFrames with dataframe split (train, test, val).
            if a size is 0.0 it will return None at the according place
        '''
        df = self._get_df(df)
        return split_train_test(test_size=test_size, val_size=val_size, 
                                stratify_col=stratify_col, df=df, col=col,
                                random_state=random_state)
        
    def split_train_test_multilabel(self, stratify_col, test_size=0.2, val_size=0.2, df=None,
                     col='img_path', uid_col='img_uid', random_state=42):
        """Splits dataframe into train, test and val datasets via multilabel stratification.
        Slow when it comes to big datasets. Use the faster split_train_test() if possible.
    
        Args:
            stratify_col (str): Column whose classes get distributed to the different splits.
                Each entry has to be a list of labels (multilabel). The entries will be merged
                according to the uid_col, to create the real multilabels for each entry.
            test_size (float): Value between 0.0 and 1.0 describing how much data is used for the test set.
                Defaults to 0.2
            val_size (float): same as test_size, but for the validation set
            df (pd.DataFrame): The data to split
            col (str): Column with samples (should be unique values if using stratify_col)
            img_uid_col (str): column saying which unique sources are in the dataframe.
                There have to be at least enough unique sources (per class in stratify_col) in this column, to place 1 instance in each split.
                Only relevant if stratify_col is not None.
                Defaults to img_uid.
            random_state (int): Seed for random operations.
                Defaults to 42
        
        Returns:
            The created splits as a tuple
        """
        df = self._get_df(df)
        return split_train_test_multilabel(stratify_col=stratify_col, test_size=test_size, val_size=val_size, df=df,
                                           col=col, uid_col=uid_col, random_state=random_state)

    def split_by_img_path(self, test_size=0.2, val_size=0.2, df=None, 
                          random_state=42):
        '''Split dataset based on img paths (for dataset with multiple 
            entries for one image)
        Args:
            test_size (float): fraction of images in df that will be used for 
                test dataset
            val_size (float): fraction of images in df that will be used for 
                test dataset
            df (pd.DataFrame): Dataframe to split
        Returns: 
            tuple: pd.DataFrames with dataframe split (train, test, val).
            if a size is 0.0 it will return None at the according place
        '''
        df = self._get_df(df)
        return split_by_img_path(test_size=test_size, val_size=val_size, df=df, 
                                 random_state=random_state)
    
    
    def split_multilabels(self, lbl_mapping, df=None, col='anno_lbl', inplace=False):
        ''' Split multilabels column into multiple single label columns
        Args:
            df (pd.DataFrame): dataframe to split
            lbl_mapping (dict): mapping from categorie to labels. 
                {categorie: [categorie lbls, ...]}
            col (str): columns containing multilabels
            inplace (bool): overwrite LOSTDataset.df with result dataframe
        Returns:
            pd.DataFrame with new columns for each categorie. Every 
                categorie has the new name col + '_' + categorie
        
        Raise:
            Exception if a multilabel contains multiple labels
                of one categorie
        '''
        df = self._get_df(df)
        df = split_multilabels(df=df, lbl_mapping=lbl_mapping, col=col)
        return self._update_inplace(df, inplace=inplace)
        
    
    
    #
    #   Validation
    #
    
    def validate_geometries(self, df=None, remove_invalid=True, inplace=False):
        '''Validate geometries 
        
        Args:
            remove_invalid (bool): remove invalid geometries from the dataframe
            df (pd.DataFrame): DataFrame to check
            inplace (bool): overwrite LOSTDataset.df with result dataframe
            
        Returns:
            None or pd.DataFrame
        '''
        df = self._get_df(df)
        df = validate_geometries(df, remove_invalid)
        return self._update_inplace(df, inplace)


    def validate_empty_images(self, df=None, inplace=False):
        '''Remove empty entries for images where non-empty entries do exist
        
        Args:
            df (pd.DataFrame): df to apply on
            inplace (bool): overwrite own dataframe with result dataframe
            
        Returns:
            pd.DataFrame: dataframe with validated empty images
        '''
        df = self._get_df(df)
        df = validate_empty_images(df)
        return self._update_inplace(df, inplace)
    
    
    def validate_unique_annos(self, df=None, inplace=False):
        '''Validate that annotations are unique
        
        Args:
            df (pd.DataFrame): df to apply on
            inplace (bool): overwrite own dataframe with result dataframe
            
        Returns:
            pd.DataFrame: dataframe with validated empty images
        '''
        df = self._get_df(df)
        df = validate_unique_annos(df)
        return self._update_inplace(df, inplace)


    def validate_image_paths(self, df=None, remove_invalid=True, inplace=False):
        '''Validate that paths do exist
        
        Args:
            remove_invalid (bool): remove non existing paths from the dataframe
            df (pd.DataFrame): df to apply on
            inplace (bool): overwrite own dataframe with result dataframe
            
        Returns:
            pd.DataFrame: dataframe with validated image paths
        '''
        df = self._get_df(df)
        df = validate_img_paths(df, remove_invalid, self.fileman)
        return self._update_inplace(df, inplace)
    
    
    def validate_single_labels(self, df=None, lbl_col='anno_lbl', 
                               dst_col='my_lbl', inplace=False):
        '''Validate if labels in a multilabel column are single labels

        Args:
            df (pd.DataFrame): Dataframe to validate single labels
            lbl_col (str): label col to validate
            ds_col (str): new column containing single label
            inplace (bool): overwrite own dataframe with result dataframe
            
        Returns:
            pd.DataFrame with additioinal column dst_col containing single labels
        '''
        df = self._get_df(df)
        df = validate_single_labels(df, lbl_col, dst_col)
        return self._update_inplace(df, inplace)
    
    
    
    #
    #   Visualization
    #
    
    def vis_and_store(self, out_dir, df=None, lbl_col='anno_lbl', 
                      color=(0, 0, 255), line_thickness='auto', 
                      fontscale='auto', radius=2, parallel=-1):
        '''Visualize annotations and store them to a folder

        Args:
            df (pd.DataFrame): Optional dataset in lost format to visualize
            out_dir (str): Directory to store the visualized annotations
            color (tuple, dict of tuple): colors (B,G,R) for all annos if tuple 
                or dict for labelwise mapping like {label: color}
            line_thickness (int, dict of int): line thickness for annotations if int
                or dict for anno-type wise mapping like {dtype: thickness}
            lbl_col (str): column containing the labels
            radius (int): radius to draw for points/circles
        '''
        df = self._get_df(df)
        vis_and_store(df, out_dir, lbl_col, color, line_thickness, fontscale, 
                      self.fileman, radius, parallel)
    
    
    def vis_semantic_segmentation(self, out_dir, n_classes, palette='dark', 
                                  seg_path_col='seg_path', df=None, parallel=-1):
        """Visualize the stored semantic segmentations by coloring it
    
        Args:
            out_dir (str): path to store images
            n_classes (int): number of classes occuring in pixelmaps, number of 
                different colors needed for visualization
            palette (str): seaborn color palette i.e. 'dark', 'bright', 'pastel',...
                refer https://seaborn.pydata.org/tutorial/color_palettes.html 
            df (pandas.DataFrame): The DataFrame that contains annoations to 
                visualize. 
        """
        df = self._get_df(df)
        vis_semantic_segmentation(df, out_dir, n_classes, palette, seg_path_col, 
                                  self.fileman, parallel=parallel)


    #
    #   Cropping
    #
    
    def crop_dataset(self, dst_dir, crop_shape=(500, 500), overlap=(0, 0), 
                     df=None, write_empty=False, fill_value=0, inplace=False, 
                     parallel=-1):
        """Crop the entire dataset with fixed crop-shape

        Args:
            df (pd.DataFrame): dataframe to apply bbox typecast
            dst_dir (str): Directory to store the new dataset
            crop_shape (tuple, list): [H, W] cropped image dimensions
            overlap (tuple, list): [H, W] overlap between crops
            write_empty (bool): Flag if crops without annos wil be written
            fill_value (float): pixel value to fill the rest of crops at borders
            inplace (bool): overwrite own dataframe with result dataframe
            
        Returns:
            pd.DataFrame
        """
        df = self._get_df(df)
        df = crop_dataset(df=df, dst_dir=dst_dir, crop_shape=crop_shape, 
                          overlap=overlap, write_empty=write_empty, 
                          fill_value=fill_value, filesystem=self.fileman, 
                          parallel=parallel)
        return self._update_inplace(df, inplace)
    
    
    def crop_anno(self, img_path, crop_position, im_w=None, im_h=None, df=None):
        """Calculate annos for a crop-position in an image
        
        Args:
            img_path (str): image to calculate the crop-annos
            crop_position (list): crop-position like: [xmin, ymin, xmax, ymax] 
                in absolute data-format
            im_w (int): width of image if available
            im_h (int): height of image if available
            df (pd.DataFrame): dataframe to apply 
            
        Returns:
            pd.DataFrame 
        """
        df = self._get_df(df)
        return self.cropper.crop_anno(img_path, df, crop_position, im_w, im_h)
    
    
    def crop_img(self, img, crop_shape=(512, 512), overlap=(0, 0), fill_value=0, 
                 df=None):
        '''Crop a 3D image and get a 4D array with crops
        
        Args:
            img (np.array): (H, W, C) Image to crop
            crop_shape (tuple): (H, W) shape of crops
            overlap (tuple): (H, W) overlap between crops
            fill_value (float): pixel-value to fill the area where crops are 
                out of the image
            df (pd.DataFrame): dataframe to apply 
        
        Returns:
            (np.ndarray, np.ndarray, tuple): 
                1. batch of cropped images as 4D-array with shape 
                    [N, crop_shape[0], crop_shape[1], channels] 
                2. positions of shape [N, 4] in format [xmin, ymin, xmax, ymax]
                3. image padding ((top_pad, bot_pad), (left_pad, right_pad))
        '''
        df = self._get_df(df)
        return self.cropper.crop_img(img, crop_shape=crop_shape, 
                                     overlap=overlap, fill_value=fill_value)
    
    
    def crop_components(self, dst_dir=None, base_labels=-1, lbl_col='anno_lbl', 
                        context=0, df=None, context_alignment=None, 
                        min_size=None, anno_dtype=['polygon'], 
                        center_lbl_key='center_lbl', inplace=False, 
                        parallel=-1):
        """Crop the entire dataset with fixed crop-shape

        Args:
            df (pd.DataFrame): dataframe to apply bbox typecast
            dst_dir (str, None): Directory to store the crops. If None no 
                cropping will be done.
            base_labels (list of str): labels to align the crops 
            lbl_col (str): column holding the labels
            context (float, tuple of floats): context to add to each component for 
                cropping (twice -> left/right, top/bottom). If tuple of float: uses 
                the given floats as fraction of the components (H, W) to calculate 
                the added contexts. 
            context_alignment (str): Define alignment of the crop-context. One of 
                [None, 'max', 'min', 'flip']. 
                None: Use H/W-context in H/W-dimension
                max: use maximum of H-context, W-context for both dimensions
                min: use minimum of H-context, W-context for both dimensions
                flip: use H-context in W-dimension and vice versa. This makes the 
                    crop a little bit more squarish shaped
            min_size (int, tuple of int): minimum size of produced crops in both
                dimensions if int or for (H, W) if tuple of int
            anno_dtype (list of str): dtype to apply on
            center_lbl_key (str): column containing the label the crop was aligned to
                
        Returns:
            pd.DataFrame
        """
        df = self._get_df(df)
        df = crop_components(df, dst_dir, base_labels, lbl_col, context, 
                             context_alignment, min_size, anno_dtype, 
                             center_lbl_key, self.fileman, parallel=parallel)
        return self._update_inplace(df, inplace)
        
    #
    #   Segmentation
    #
    
    def semantic_segmentation(self, order, dst_dir, fill_value, df=None, 
                              anno_dtypes=['polygon'], use_empty=False, 
                              lbl_col='anno_lbl', dst_path_col='seg_path', 
                              dst_lbl_col='seg_lbl', line_thickness=None, 
                              radius=None, numeric_filename=False, 
                              inplace=False, parallel=-1):
        '''Create semantic segmentations from polygon-annos
    
        Args:
            df (pd.DataFrame): dataframe to generate the pixel maps from
            order (dict, list): order of the classes. Higher pixel
                values will overwrite lower ones. 
                Example: order=['Car', 'Person', 'Glasses'] or  
                    order={'Car': 0, 'Person': 1, 'Glasses': 2}
                    Car will get pixel value 0, Person px value 1 and so on - 
                    Person overwrites Car, Glasses overwrites Person ... 
            dst_dir (str): Directory to store the pixel maps.
            fill_value (int): Pixel value for not annotated areas. Usually this
                will be something like 0 for something like 'background'
            anno_dtypes (list of string): dtypes to use for segmentation. Possible
                values are {'polygon', 'bbox', 'line', 'point'}. Annotations with
                dtypes other than anno_dtypes will be removed from dataframe
            lbl_col (str): Column containing the training labels
            line_thickness (int): thickness of line-segmentation when using an 
                annotation of dtype line. Only takes effect when anno_dtypes 
                contains 'line'
            radius (int): radius of circle-segmentation when using an annotation
                of dtype point. Only takes effect when anno_dtypes contains 
                'point'
            numeric_filename (bool): use same filename for seg and img if False. If
                True a numeric 12-digit filename will be generated. Usefull if some 
                files have same filenames.
            inplace (bool): overwrite self.df with result df
                
        Returns:
            pd.DataFrame: The original dataframe with new column (dst_col) 
                containing the path to the according segmentation file. 
                Furthermore the column dst_lbl_col contains the label the
                segmentation looked up in order for creation
        '''
        df = self._get_df(df)
        df = semantic_segmentation(order, dst_dir, fill_value, df, anno_dtypes,
                                   use_empty, lbl_col, dst_path_col, 
                                   dst_lbl_col, line_thickness, radius, 
                                   self.fileman, numeric_filename, 
                                   parallel=parallel)
        return self._update_inplace(df, inplace)
    
    
    def segmentation_to_lost(self, pixel_mapping, background=0, 
                             seg_key='seg_path', df=None, cast_others=False, 
                             inplace=False, parallel=-1):
        '''Create LOST-Annotations from semantic segmentations / pixelmaps 
        
        Args:
            pixel_mapping (dict): mapping of pixel-value to anno_lbl, e.g. 
                {0:'background', 1:'thing', 255:'another_thing'}. 
                If no annotations for background are desired it can be ommited 
                from the dictionary
            background (int): pixel-value for background.
            seg_key (str): dataframe key containing the paths to the stored 
                pixelmaps
            cast_others (bool): Flag if pixel values not occuring in 
                `pixel_mapping` should be handeled as background. Raises 
                ValueError if False and an unknown pixel occurs
            df (pd.DataFrame): dataframe to apply on
            
        Returns:
            pd.DataFrame with polygon-annotations in LOSTDataset format
        '''
        df = self._get_df(df)
        df = segmentation_to_lost(df, pixel_mapping=pixel_mapping, 
                                      background=background, seg_key=seg_key, 
                                      cast_others=cast_others,
                                      filesystem=self.fileman, 
                                      parallel=parallel)
        return self._update_inplace(df, inplace)
    
    
    #
    #   Detection
    #
    
    def detection_dataset(self, df=None, lbl_col='anno_lbl', det_col='det_lbl', 
                          bbox_style='x1y1x2y2', use_empty_images=False, 
                          inplace=False):
        '''Prepare all bboxes to use them for detection CNN training
        Args:
            df (pd.DataFrame): Dataframe containing bbox annotations
            lbl_col (str): column name where the anno labels are located (single 
                label or multilabel)
            det_col (str): column name where the training labels are located (single 
                label only)
            bbox_style (str): bbox anno-style {'xywh', 'x1y1x2y2', 'xcycwh'}
            use_empty_images (bool, str, int): specifiy usage of empty images 
                (empty image means image without bbox annotation).
                True: keep all images, empty and non-empty
                False: only keep non-empty images and drop all empty images
                'balanced': If more empty images than non-empty ones do exist a 
                    random selection will be sampled to have the same amount of 
                    empty and non-empty images. If less empty than non-empty images 
                    do exist all of them will be kept
                int: a specific amount of empty images will be samples randomly
            inplace (bool): overwrite self.df with result df
                
        Returns:
            pd.DataFrame: detection dataset

        Note:
            Other anno-types than 'bbox' will be ignored. You can transform 
            polygons to bboxes before by calling LOSTDataset.polygon_to_bbox()
        '''
        df = self._get_df(df)
        df = detection_dataset(df, lbl_col, det_col, bbox_style, 
                               use_empty_images, self.fileman)
        return self._update_inplace(df, inplace)
    
    
    def bbox_nms(self, lbl_col='anno_lbl', method='nms', iou_thr=0.5, df=None, 
                 inplace=False, parallel=-1, **kwargs):
        """apply nms to dataframe to suppress overlapping bboxes

        Args:
            df (pd.DataFrame): dataframe containing bboxes to supress
            lbl_col (str, optional): columns containing labels. Defaults to 'anno_lbl'.
            method (str, optional): one of {'nms', 'nmw', 'soft_nms', 'wbf', 'merge'}. Defaults to 'nms'.
            iou_thr (float, optional): iou threshold when to supress bboxes. Defaults to 0.5.
            filesystem (fsspec, optional): filesystem. Defaults to None.
            parallel (int): use parallelism

        Returns:
            pd.DataFrame: dataframe with overlapping bboxes pruned
        """
        df = self._get_df(df)
        df = bbox_nms(df, lbl_col, method, iou_thr, self.fileman, parallel, 
                      **kwargs)
        return self._update_inplace(df, inplace)
        
    
    #
    #   Masking
    #
    
    def mask_dataset(self, dst_dir, masking_labels, mask_value=0, inverse=False, 
                     lbl_col='anno_lbl', dst_col='img_path', df=None, 
                     inplace=False, parallel=-1):
        """mask images by annos of given label. Either mask the given annos itself 
        (inverse=False) or everything but the annos (inverse=True)

        Args:
            dst_dir (str): destination folder
            masking_labels (list): list of labels to use for masking
            mask_value (int, optional): pixel value to use for mask. Defaults to 0.
            inverse (bool, optional): fill the given annos with `mask_value` if True, and inverse if False. Defaults to False.
            lbl_col (str, optional): column containing the labels. Defaults to 'anno_lbl'.
            df (pd.DataFrame): dataframe to mask
            inplace (bool): overwrite self.df with result df
        """
        df = self._get_df(df)
        df = mask_dataset(df, dst_dir, masking_labels, mask_value, inverse, 
                          lbl_col, dst_col, self.fileman, parallel=parallel)
        return self._update_inplace(df, inplace)


import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split

from lost_ds.functional.filter import img_selection, label_selection
from iterstrat.ml_stratifiers import MultilabelStratifiedShuffleSplit
from sklearn.preprocessing import MultiLabelBinarizer


def split_by_empty(df, col='anno_data'):
    '''Split dataframe into empty and non-empty entries, filtered by given col
    Args:
        df (pd.DataFrame): df to apply on
        col (str): column to check for empty entries
    Returns:
        tuple of pd.DataFrame: first dataframe contains non-emtpy entries, 
            the second dataframe contains empty entries
    '''
    empty_bool = df[col].isnull()
    empty_df = df[empty_bool]
    not_empty_df = df[~empty_bool]
    return not_empty_df, empty_df

def split_train_test(test_size=0.2, val_size=0.2, stratify_col=None, df=None,
                     col='img_path', uid_col=None, random_state=42):
    """Splits dataframe into train, test and val datasets.
    
    Args:
        test_size (float): Value between 0.0 and 1.0 describing how much data is used for the test set.
            Defaults to 0.2
        val_size (float): same as test_size, but for the validation set
        stratify_col (str): Column whose classes get distributed to the different splits if this arg is not None.
            Defaults to None
        df (pd.DataFrame): The data to split
        col (str): Column with samples (should be unique values if using stratify_col)
        img_uid_col (str): column saying which unique sources are in the dataframe.
            There have to be at least 3 unique sources (per class in stratify_col) in this column, to make the 3 splits (fairly).
            Only relevant if stratify_col is not None.
            Defaults to img_uid.
        random_state (int): Seed for random operations.
            Defaults to 42
    
    Returns:
        The created splits as a tuple
    """
    df_base = df.copy()
    if not len(df_base):
        return df_base, df_base, df_base
    df_base['split_col'] = df_base[col].map(lambda x: hash(str(x)))
    samples = df_base['split_col'].unique()
    n = len(samples)
    stratify = None
    ids = list(range(n))

    nr_splits = 3
    if not test_size:
        nr_splits-=1
    if not val_size:
        nr_splits-=1

    # validaton + filtering for stratification
    if stratify_col:
        stratify = list(df_base[stratify_col])
        assert len(samples) == len(df_base), 'Samples cannot occur multiple times in dataset when using stratify!'

        if uid_col is not None:
            unique_classes = list(df_base[stratify_col].unique())
            # check nr of unique sources per class
            # HACK: could use validate_single_labels() here, but that causes circular imports...
            for u_class in unique_classes:
                class_df = df_base[df_base[stratify_col] == u_class]
                unique_imgs = class_df[uid_col].unique()
                nr_unique_imgs = len(unique_imgs)
                print(u_class, nr_unique_imgs)
                # has enough base-imgs per class per split
                if nr_unique_imgs < nr_splits:
                    to_drop = df_base[df_base[uid_col].isin(unique_imgs)]
                    len_to_drop = len(to_drop)
                    df_base = df_base[~df_base[uid_col].isin(unique_imgs)]
                    print(f"""Dropped {len_to_drop} entries based off of {nr_unique_imgs} images of class {u_class}, 
                        due to it not having enough unique source-images""")
            samples = df_base['split_col'].unique()
            n = len(samples)
            ids = list(range(n))
            stratify = list(df_base[stratify_col])
    
    # doing the splitting
    splits = []
    for split in [test_size, val_size]:
        if split:
            size = int(split * n)
            # something left to split? (return empty df_base if not)
            if size == 0:
                return df_base, df_base, df_base
            set_1, set_2, ids_1, ids_2 = train_test_split(samples, ids, test_size=size, 
                                                          shuffle=True, 
                                                          random_state=random_state,
                                                          stratify=stratify)
            split_data = label_selection(list(set_2), df=df_base, col='split_col')
            splits.append(split_data)
            samples = set_1
            ids = ids_1
            if stratify_col:
                stratify = list(df_base.iloc[ids_1][stratify_col])
        else:
            splits.append(None)
    train_data = label_selection(list(samples), df=df_base, col='split_col')
    splits.insert(0, train_data)
    return tuple(splits)


def create_multilabel_data(df, anno_col='anno_lbl', mult_col='mult_lbl', single_lbls='single_lbl'):
    """Creates a column with multilabel data and drops unique image-paths"""
    unique_img_df = df.drop_duplicates('img_path')
    for path, path_df in df.groupby('img_path'):
        multi_lbl_list = path_df[single_lbls].unique()
        indexes = path_df.index
        unique_img_df.at[indexes[0], mult_col] = multi_lbl_list
        
    return unique_img_df


def split_train_test_multilabel(stratify_col, test_size=0.2, val_size=0.2, df=None,
                     col='img_path', uid_col='img_uid', random_state=42):
    """Splits dataframe into train, test and val datasets via multilabel stratification.
        Slow when it comes to big datasets. Use the faster split_train_test() if possible.
    
    Args:
        stratify_col (str): Column whose classes get distributed to the different splits.
            Each entry has to be a list of labels (multilabel). The entries will be merged
            according to the uid_col, to create the real multilabels for each entry.
        test_size (float): Value between 0.0 and 1.0 describing how much data is used for the test set.
            Defaults to 0.2
        val_size (float): same as test_size, but for the validation set
        df (pd.DataFrame): The data to split
        col (str): Column with samples (should be unique values if using stratify_col)
        img_uid_col (str): column saying which unique sources are in the dataframe.
            There have to be at least enough unique sources (per class in stratify_col) in this column, to place 1 instance in each split.
            Only relevant if stratify_col is not None.
            Defaults to img_uid.
        random_state (int): Seed for random operations.
            Defaults to 42
    
    Returns:
        The created splits as a tuple
    """
    from lost_ds.functional.validation import validate_single_labels # HACK: prevent circular import
    df_base = df.copy()

    # check nr of splits
    nr_splits = 3
    if not test_size:
        nr_splits-=1
    if not val_size:
        nr_splits-=1
    # check nr of unique sources per class; drop if less than nr of splits
    df_base = validate_single_labels(df_base, lbl_col=stratify_col, dst_col='single_lbl')
    unique_img_lbl_df = df_base.drop_duplicates(subset=[uid_col,'single_lbl'])
    lbl_count = unique_img_lbl_df['single_lbl'].value_counts()
    for clss in lbl_count.keys():
        if lbl_count[clss] < nr_splits:
            to_drop_uids = df_base[df_base['single_lbl'] == clss][uid_col]
            df_base = df_base[~df_base[uid_col].isin(to_drop_uids)]
            print(f"Dropped {len(to_drop_uids)} entries of class {clss}")

    multi_lbl_df = create_multilabel_data(df_base, anno_col=stratify_col)
    multi_lbl_df = multi_lbl_df.reset_index(drop=True)
    if not len(multi_lbl_df):
        print("Return empty splits")
        return multi_lbl_df, multi_lbl_df, multi_lbl_df
    multi_lbl_df['split_col'] = multi_lbl_df[col].map(lambda x: hash(str(x)))
    samples = multi_lbl_df['split_col'].unique()    
    n = len(samples)
    ids = np.array(list(range(n)))
    assert len(samples) == len(multi_lbl_df), 'Samples cannot occur multiple times in dataset when using stratify!'

    # doing the splitting
    split_paths = []
    multilabels = multi_lbl_df[stratify_col] # new labels after dropping
    mult_lbl_binarizer = MultiLabelBinarizer()
    hot_encoding = mult_lbl_binarizer.fit_transform(multilabels)
    for split in [test_size, val_size]:
        if split:
            size = int(split * n)
            # something left to split? (return empty base_df if not)
            if size == 0:
                return df_base, df_base, df_base
            multisplitter = MultilabelStratifiedShuffleSplit(test_size=size, random_state=random_state, n_splits=1)
            for train_index, test_index in multisplitter.split(samples, hot_encoding):
                set_1, set_2 = samples[train_index], samples[test_index]
                ids_1, ids_2 = ids[train_index], ids[test_index]
                hot_encoding_1, hot_encoding_2 = hot_encoding[train_index], hot_encoding[test_index]
            split_data = label_selection(list(set_2), df=multi_lbl_df, col='split_col')
            split_paths.append(split_data['img_path'])
            samples = set_1
            ids = ids_1
            hot_encoding = hot_encoding_1
        else:
            split_paths.append(None)
    train_data = label_selection(list(samples), df=multi_lbl_df, col='split_col')
    split_paths.insert(0, train_data['img_path'])
    split_dfs = []
    for paths in split_paths:
        if paths is not None:
            split_df = df_base[df_base['img_path'].isin(paths)]
            split_dfs.append(split_df)
        else:
            split_dfs.append(None)
    
    return tuple(split_dfs)

def split_by_img_path(test_size=0.2, val_size=0.2, df=None, random_state=42):
    '''Split dataset based on img paths (for dataset with multiple 
        entries for one image)
    Args:
        test_size (float): fraction of images in df that will be used for 
            test dataset
        val_size (float): fraction of images in df that will be used for 
            test dataset
        df (pd.DataFrame): Dataframe to split
    Returns: 
        tuple: pd.DataFrames with dataframe split (train, test, val).
        if a size is 0.0 it will return None at the according place
    '''
    return split_train_test(test_size, val_size, None, df, 'img_path', 
                            random_state)


def split_multilabels(lbl_mapping, df:pd.DataFrame=None, col='anno_lbl'):
    ''' Split multilabels column into multiple single label columns
    Args:
        df (pd.DataFrame): dataframe to split
        lbl_mapping (dict): mapping from categorie to labels. 
            {categorie: [categorie lbls, ...]}
        col (str): columns containing multilabels
    Returns:
        pd.DataFrame with new columns for each categorie. Every 
            categorie has the new name col + '_' + categorie
    
    Raise:
        Exception if a multilabel contains multiple labels
            of one categorie
    '''
    df = df.copy()
    cat_to_lbl = {col+'_'+cat: lbl for cat, lbl in lbl_mapping.items()}
    lbl_to_cat = {lbl: cat for cat, lbls in cat_to_lbl.items() for lbl in lbls}
    categories = list(cat_to_lbl.keys())
    df[categories] = None
    ret_dict = {cat: None for cat in categories}
    
    def split_labels(labels):
        ret = ret_dict.copy()
        for lbl in labels:
            if lbl is not None:
                cat = lbl_to_cat[lbl]
                if ret[cat] is not None:
                    raise Exception('Got multilabel for categorie {}. ' \
                                    'Found {}'.format(cat, ret[cat], lbl))
                ret[cat] = lbl
        ret = list(ret.values())
        for lbl in labels:
            assert lbl in ret
        return ret
    
    df[categories] = pd.DataFrame(list(df[col].map(
        lambda x: split_labels(x))), columns=categories, index=df.index)
    
    # TODO: continue implementation
    # df[categories] = df[col].map(split_labels, axis=1, result_type='expand')
    
    return df


import os 
import numpy as np 
from joblib import Parallel, delayed 
from tqdm import tqdm 

from lost_ds.functional.filter import label_selection, unique_labels
from lost_ds.geometry.lost_geom import LOSTGeometries
from lost_ds.util import get_fs



def mask_dataset(df, dst_dir, masking_labels, mask_value=0, inverse=False, 
                 lbl_col='anno_lbl', dst_col='img_path', filesystem=None, 
                 parallel=-1):
    """mask images by annos of given label. Either mask the given annos itself 
    (inverse=False) or everything but the annos (inverse=True)

    Args:
        df (pd.DataFrame): dataframe to mask
        dst_dir (str): destination folder
        masking_labels (list): list of labels to use for masking. Use all labels if 'all' or -1
        mask_value (int, optional): pixel value to use for mask. Defaults to 0.
        inverse (bool, optional): fill the given annos with `mask_value` if True, and inverse if False. Defaults to False.
        lbl_col (str, optional): column name containing labels. Defaults to 'anno_lbl'.
        mask_path (str, optional): new column name for masked img paths. Defaults to 'img_path'
        filesystem (_type_, optional): _description_. Defaults to None.
    """
    fs = get_fs(filesystem)
    fs.makedirs(dst_dir, True)
    df = df.copy()
    # TODO: only polygon and bbox can be masked
    
    if masking_labels == -1 or masking_labels == 'all':
        masking_labels = unique_labels(df, lbl_col)
    
    def _mask_img(path, annos):
        mask_annos = label_selection(masking_labels, annos, lbl_col)
        dst_path = os.path.join(dst_dir, path.split('/')[-1])
        
        # mask labels occured -> apply mask to image
        if len(mask_annos):    
            img = fs.read_img(path)
            h, w = img.shape[:2]
            c = img.shape[-1] if len(img.shape) == 3 else 1
            mask = np.zeros((h, w))
            geom = LOSTGeometries()
            set_val = 1
            for _, row in mask_annos.iterrows():
                mask = geom.segmentation(mask, set_val, row.anno_data, 
                                        row.anno_dtype, row.anno_format, 
                                        row.anno_style)
            if inverse:
                set_val = 0
            px = (mask_value, )*3 if c==3 else mask_value
            img = np.where(mask[..., None]==set_val, px, img)
            fs.write_img(img, dst_path)
        
        # no mask labels occured -> copy image or write mask-only if invers
        else:
            if inverse:
                fs.write_img(np.full_like(fs.get_imagesize(path), mask_value), 
                             dst_path)
            else:
                fs.copy(path, dst_path)
    
    if parallel:
        Parallel(n_jobs=parallel)(delayed(_mask_img)(path, annos) 
                    for path, annos in tqdm(df.groupby('img_path'), 
                                            desc='mask imgs')
                    )
    else:
        for path, annos in tqdm(df.groupby('img_path'), desc='mask imgs'):
            _mask_img(path, annos)
    
    
    
    # remap new path
    df[dst_col] = df['img_path'].map(lambda x: 
        os.path.join(dst_dir, x.split('/')[-1]))
    
    return df 





import cv2
from tqdm import tqdm
from joblib import Parallel, delayed
import numpy as np 
import pandas as pd 

from lost_ds.im_util import get_fs


def _segmentation_to_polygon(segmentation, pixel_mapping: dict, background, 
                             cast_others):
    if len(segmentation.shape)==3:
        segmentation = cv2.cvtColor(segmentation, cv2.COLOR_BGR2GRAY)
    u_values = np.unique(segmentation)
    expected_values = np.unique(list(pixel_mapping.keys()) + [background])
    setdiff = np.setdiff1d(u_values, expected_values)
    if len(setdiff) and not cast_others:
        raise ValueError(f'Got pixel values {u_values} expected ' \
                         f'{expected_values} from {pixel_mapping} ' \
                         f'(unexpected values {list(setdiff)})')
    
    im_h, im_w = segmentation.shape
    
    lbl_name = []
    lost_polygons = []
    
    # iterate over all classes and get their segmentations
    for v in u_values:
        if not v in pixel_mapping.keys():
            continue
        # create color pixel map to detect contours of class v
        v_seg = segmentation.copy()
        v_seg[segmentation!=v] = 0
        v_seg[segmentation==v] = 255
        
        # get contours
        contours, _ = cv2.findContours(v_seg, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE)
        contours = map(np.squeeze, contours)
        cont_seg = np.zeros_like(v_seg)
        valid_contours, contour_pixels = list(), list()
        
        # validate contours
        for idx, cont in enumerate(contours):
            if len(cont) < 4:
                continue
            cont_seg = cv2.drawContours(cont_seg, [cont], 0, idx + 1, -1)# fill poly
            cont_seg = cv2.drawContours(cont_seg, [cont], 0, 0, 1) # ignore border
            valid_contours.append(cont)
            contour_pixels.append(idx + 1)
                
        for px_pos, cont in zip(contour_pixels, valid_contours):
            # pick pixels
            px_vals = segmentation[cont_seg==px_pos]
            # find unique pixels
            u_px_vals, cnts = np.unique(px_vals, return_counts=True)
            if not len(u_px_vals):
                continue
            # choose most occuring pixel as class
            px = u_px_vals[np.argmax(cnts)]
            if px != v:
                continue
            # consider unknown pixel value as background if `cast_others`
            if not px in expected_values and cast_others:
                px = 0
            # ignore background if not provided
            if px == background and not background in list(pixel_mapping.keys()):
                continue
            cont = np.array(cont) / np.array([im_w, im_h])
            lost_polygons.append(cont)
            lbl_name.append([pixel_mapping[px]])

    return lost_polygons, lbl_name

def segmentation_to_lost(df: pd.DataFrame, pixel_mapping, background=0, 
                         seg_key='seg_path', cast_others=False, filesystem=None,
                         parallel=-1):
    
    '''Create LOST-Annotations from semantic segmentations / pixelmaps 
    
    Args:
        pixel_mapping (dict): mapping of pixel-value to anno_lbl, e.g. 
            {0:'background', 1:'thing'}
        background (int): pixel-value for background.
        seg_key (str): dataframe key containing the paths to the stored 
            pixelmaps
        df (pd.DataFrame): dataframe to apply on
        cast_others (bool): Flag if pixel values not occuring in `pixel_mapping` 
            should be handeled as background. Raises ValueError if False and 
            an unknown pixel occurs
        filesystem (fsspec.filesystem, FileMan): filesystem to use. Use local
            if not initialized
        
    Returns:
        pd.DataFrame with polygon-annotations in LOSTDataset format
    '''
    
    fs = get_fs(filesystem)
    
    pixel_mapping = pixel_mapping.copy()
    if background != 0:
        raise NotImplementedError('Background pixels have to be 0')
    
    def seg_to_poly(seg_path, seg_df):
        img_paths = seg_df['img_path'].unique()
        assert len(img_paths) == 1
        lost_anno_data = dict()
        segmentation = fs.read_img(seg_path, 0)            
        lost_polys, lbl_name = _segmentation_to_polygon(segmentation, 
                                                        pixel_mapping, 
                                                        background,
                                                        cast_others)
        assert len(lost_polys) == len(lbl_name)
        img_path = img_paths[0]
        lost_anno_data['anno_lbl'] = lbl_name
        lost_anno_data['anno_data'] = lost_polys
        lost_anno_data['img_path'] = [img_path] * len(lbl_name)
        lost_anno_data[seg_key] = [seg_path] * len(lbl_name)
        lost_anno_data['anno_format'] = ['rel'] * len(lbl_name)
        lost_anno_data['anno_style'] = ['xy'] * len(lbl_name)
        lost_anno_data['anno_dtype'] = ['polygon'] * len(lbl_name)
        return pd.DataFrame(lost_anno_data)
    
    n_segs = len(df[seg_key].unique())
    if parallel:
        ret = Parallel(n_jobs=parallel)(delayed(seg_to_poly)(seg_path, seg_df) 
                                for seg_path, seg_df in tqdm(df.groupby([seg_key]), 
                                                            total=n_segs, 
                                                            desc='seg. to anno'))
    else:
        ret = list()
        for seg_path, seg_df in tqdm(df.groupby([seg_key])):
            ret.append(seg_to_poly(seg_path, seg_df))
        
    df = pd.concat(ret)
    df.reset_index(inplace=True, drop=True)
    
    return df

